import torch

import numpy as np
import torch.nn.functional as F
import torch.nn as nn
import datetime
import random
from sklearn.metrics import roc_auc_score, confusion_matrix, precision_recall_curve, auc, \
    mean_absolute_error, r2_score



def statistical(y_true, y_pred, y_pro):
    c_mat = confusion_matrix(y_true, y_pred)
    tn, fp, fn, tp = list(c_mat.flatten())
    se = tp/(tp+fn)
    sp = tn/(tn+fp)
    acc = (tp+tn)/(tn+fp+fn+tp)
    mcc = (tp*tn-fp*fn)/np.sqrt((tp+fp)*(tp+fn)*(tn+fp)*(tn+fn)+1e-8)
    auc_prc = auc(precision_recall_curve(y_true, y_pro, pos_label=1)[1],
                  precision_recall_curve(y_true, y_pro, pos_label=1)[0])
    auc_roc = roc_auc_score(y_true, y_pro)
    return tn, fp, fn, tp, se, sp, acc, mcc, auc_prc, auc_roc


class Meter(object):
    """Track and summarize model performance on a dataset for
    (multi-label) binary classification."""
    def __init__(self):
        self.mask = []
        self.y_pred = []
        self.y_true = []

    def update(self, y_pred, y_true, mask):
        """Update for the result of an iteration

        Parameters
        ----------
        y_pred : float32 tensor
            Predicted molecule labels with shape (B, T),
            B for batch size and T for the number of tasks
        y_true : float32 tensor
            Ground truth molecule labels with shape (B, T)
        mask : float32 tensor
            Mask for indicating the existence of ground
            truth labels with shape (B, T)
        """
        self.y_pred.append(y_pred.detach().cpu())
        self.y_true.append(y_true.detach().cpu())
        self.mask.append(mask.detach().cpu())

    def roc_precision_recall_score(self):
        """Compute AUC_PRC for each task.
        Returns
        -------
        list of float
            rmse for all tasks
        """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_pred = torch.sigmoid(y_pred)
        y_true = torch.cat(self.y_true, dim=0)
        n_data, n_tasks = y_true.shape
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()
            precision, recall, _thresholds = precision_recall_curve(task_y_true, task_y_pred, pos_label=1)
            scores.append(auc(recall, precision))

        return scores

    def roc_auc_score(self):
        """Compute roc-auc score for each task.

        Returns
        -------
        list of float
            roc-auc score for all tasks
        """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        # Todo: support categorical classes
        # This assumes binary case only
        y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率
        n_tasks = y_true.shape[1]

        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()
            scores.append(roc_auc_score(task_y_true, task_y_pred))

        return scores
    def se(self):
        """Compute se score for each task.

                Returns
                -------
                list of float
                    se score for all tasks
                """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        # Todo: support categorical classes
        # This assumes binary case only
        y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率
        n_tasks = y_true.shape[1]
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()

            task_y_pred = [1 if i >= 0.5 else 0 for i in task_y_pred]
            c_mat = confusion_matrix(task_y_true, task_y_pred)

            tn, fp, fn, tp = list(c_mat.flatten())
            se = tp / (tp + fn)
            scores.append(se)

        return scores
    def precision(self):
        """Compute precision score for each task.

                Returns
                -------
                list of float
                    precision score for all tasks
                """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        # Todo: support categorical classes
        # This assumes binary case only
        y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率
        n_tasks = y_true.shape[1]
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()

            task_y_pred = [1 if i >=0.5 else 0 for i in task_y_pred]
            c_mat = confusion_matrix(task_y_true, task_y_pred)

            tn, fp, fn, tp = list(c_mat.flatten())
            precision = tp / (tp + fp)
            scores.append(precision)

        return scores
    def sp(self):
        """Compute sp score for each task.

                Returns
                -------
                list of float
                    sp score for all tasks
                """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        # Todo: support categorical classes
        # This assumes binary case only
        y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率
        n_tasks = y_true.shape[1]
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()

            task_y_pred = [1 if i >=0.5 else 0 for i in task_y_pred]
            c_mat = confusion_matrix(task_y_true, task_y_pred)

            tn, fp, fn, tp = list(c_mat.flatten())

            sp = tn / (tn + fp)

            scores.append(sp)

        return scores
    def acc(self):
        """Compute acc score for each task.

                Returns
                -------
                list of float
                    acc score for all tasks
                """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        # Todo: support categorical classes
        # This assumes binary case only
        y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率
        n_tasks = y_true.shape[1]
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()

            task_y_pred = [1 if i >=0.5 else 0 for i in task_y_pred]
            c_mat = confusion_matrix(task_y_true, task_y_pred)

            tn, fp, fn, tp = list(c_mat.flatten())

            acc = (tp + tn) / (tn + fp + fn + tp)

            scores.append(acc)

        return scores
    def mcc(self):
        """Compute mcc score for each task.

                Returns
                -------
                list of float
                    mcc score for all tasks
                """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        # Todo: support categorical classes
        # This assumes binary case only
        y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率
        n_tasks = y_true.shape[1]
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()
            task_y_pred = [1 if i >=0.5 else 0 for i in task_y_pred]
            c_mat = confusion_matrix(task_y_true, task_y_pred)
            tn, fp, fn, tp = list(c_mat.flatten())
            mcc = (tp * tn - fp * fn) / np.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn) + 1e-8)
            scores.append(mcc)

        return scores
    # def r2(self):
    #     """Compute r2 score for each task.
    #
    #     Returns
    #     -------
    #     list of float
    #         r2 score for all tasks
    #     """
    #     mask = torch.cat(self.mask, dim=0)
    #     y_pred = torch.cat(self.y_pred, dim=0)
    #     y_true = torch.cat(self.y_true, dim=0)
    #     n_data, n_tasks = y_true.shape
    #     scores = []
    #     for task in range(n_tasks):
    #         task_w = mask[:, task]
    #         task_y_true = y_true[:, task][task_w != 0]
    #         task_y_pred = y_pred[:, task][task_w != 0]
    #         scores.append(r2_score(task_y_true, task_y_pred))
    #     return scores





    def l1_loss(self, reduction):
        """Compute l1 loss for each task.

        Returns
        -------
        list of float
            l1 loss for all tasks
        reduction : str
            * 'mean': average the metric over all labeled data points for each task
            * 'sum': sum the metric over all labeled data points for each task
        """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        n_tasks = y_true.shape[1]
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0]
            task_y_pred = y_pred[:, task][task_w != 0]
            scores.append(F.l1_loss(task_y_true, task_y_pred, reduction=reduction).item())
        return scores

    def rmse(self):
        """Compute RMSE for each task.

        Returns
        -------
        list of float
            rmse for all tasks
        """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        n_data, n_tasks = y_true.shape
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0]
            task_y_pred = y_pred[:, task][task_w != 0]
            scores.append(np.sqrt(F.mse_loss(task_y_pred, task_y_true).cpu().item()))
        return scores

    def mae(self):
        """Compute mae for each task.

        Returns
        -------
        list of float
            mae for all tasks
        """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        n_data, n_tasks = y_true.shape
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0]
            task_y_pred = y_pred[:, task][task_w != 0]
            scores.append(mean_absolute_error(task_y_true, task_y_pred))
        return scores

    def r2(self):
        """Compute r2 score for each task.

        Returns
        -------
        list of float
            r2 score for all tasks
        """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        n_data, n_tasks = y_true.shape
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0]
            task_y_pred = y_pred[:, task][task_w != 0]
            scores.append(r2_score(task_y_true, task_y_pred))
        return scores

    def compute_metric(self, metric_name, reduction='mean'):
        """Compute metric for each task.

        Parameters
        ----------
        metric_name : str
            Name for the metric to compute.
        reduction : str
            Only comes into effect when the metric_name is l1_loss.
            * 'mean': average the metric over all labeled data points for each task
            * 'sum': sum the metric over all labeled data points for each task

        Returns
        -------
        list of float
            Metric value for each task
        """
        assert metric_name in ['roc_auc', 'l1', 'rmse', 'prc_auc', 'mae', 'r2', 'se', 'sp', 'acc', 'mcc', 'pred',
                               'precision'], \
            'Expect metric name to be "roc_auc", "l1", "rmse", "prc_auc", "mae", "r2","pred" got {}'.format(
                metric_name)  # assert（断言）用于判断一个表达式，在表达式条件为 false 的时候触发异常
        assert reduction in ['mean', 'sum']
        if metric_name == 'roc_auc':
            return self.roc_auc_score()
        if metric_name == 'l1':
            return self.l1_loss(reduction)
        if metric_name == 'rmse':
            return self.rmse()
        if metric_name == 'prc_auc':
            return self.roc_precision_recall_score()
        if metric_name == 'mae':
            return self.mae()
        if metric_name == 'r2':
            return self.r2()
        if metric_name == 'mcc':
            return self.mcc()
        if metric_name == 'se':
            return self.se()

        if metric_name == 'precision':
            return self.precision()
        if metric_name == 'sp':
            return self.sp()
        if metric_name == 'acc':
            return self.acc()
        if metric_name == 'mcc':
            return self.mcc()


class MyDataset(object):
    def __init__(self, Xs, Ys):
        self.Xs = torch.tensor(Xs, dtype=torch.float32)
        self.masks = torch.tensor(~np.isnan(Ys) * 1.0, dtype=torch.float32)
        # convert np.nan to 0
        self.Ys = torch.tensor(np.nan_to_num(Ys), dtype=torch.float32)


    def __len__(self):
        return len(self.Ys)

    def __getitem__(self, idx):

        X = self.Xs[idx]
        Y = self.Ys[idx]
        mask = self.masks[idx]

        return X, Y, mask


class EarlyStopping(object):
    """Early stop performing

    Parameters
    ----------
    mode : str
        * 'higher': Higher metric suggests a better model
        * 'lower': Lower metric suggests a better model
    patience : int
        Number of epochs to wait before early stop
        if the metric stops getting improved
    filename : str or None
        Filename for storing the model checkpoint
    """
    def __init__(self, mode='higher', patience=10, filename=None):  # 可以加tolerance
        if filename is None:
            dt = datetime.datetime.now()
            filename = '{}_early_stop_{}_{:02d}-{:02d}-{:02d}.pth'.format(
                dt.date(), dt.hour, dt.minute, dt.second)

        assert mode in ['higher', 'lower']
        self.mode = mode
        if self.mode == 'higher':
            self._check = self._check_higher  # 私有变量（private），只有内部可以访问，外部不能访问
        else:
            self._check = self._check_lower

        self.patience = patience
        self.counter = 0
        self.filename = filename
        self.best_score = None
        self.early_stop = False

    def _check_higher(self, score, prev_best_score):
        return (score > prev_best_score)

    def _check_lower(self, score, prev_best_score):
        return (score < prev_best_score)

    def step(self, score, model):
        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(model)
        elif self._check(score, self.best_score):  # 当前模型如果是更优模型，则保存当前模型
            self.best_score = score
            self.save_checkpoint(model)
            self.counter = 0
        else:
            self.counter += 1
            # print(
            #     f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
        return self.early_stop

    def save_checkpoint(self, model):
        '''Saves model when the metric on the validation set gets improved.'''
        torch.save({'model_state_dict': model.state_dict()}, self.filename)

    def load_checkpoint(self, model):
        '''Load model saved with early stopping.'''
        model.load_state_dict(torch.load(self.filename)['model_state_dict'])


class MyDNN(nn.Module):
    def __init__(self, inputs, hideen_units, outputs, dp_ratio, reg):
        """
        :param inputs: number of inputs
        :param hideen_units: [128, 256, 512]
        :param out_puts: number of outputs
        :param dp_ratio:
        :param reg:
        """
        super(MyDNN, self).__init__()
        # parameters
        self.reg = reg

        # layers
        self.hidden1 = nn.Linear(inputs, hideen_units[0])
        self.dropout1 = nn.Dropout(dp_ratio)

        self.hidden2 = nn.Linear(hideen_units[0], hideen_units[1])
        self.dropout2 = nn.Dropout(dp_ratio)

        self.hidden3 = nn.Linear(hideen_units[1], hideen_units[2])
        self.dropout3 = nn.Dropout(dp_ratio)

        if reg:
            self.output = nn.Linear(hideen_units[2], 1)
        else:
            self.output = nn.Linear(hideen_units[2], outputs)

    def forward(self, x):
        x = self.hidden1(x)
        x = F.relu(self.dropout1(x))

        x = self.hidden2(x)
        x = F.relu(self.dropout2(x))

        x = self.hidden3(x)
        x = F.relu(self.dropout3(x))

        return self.output(x)


def collate_fn(data_batch):
    Xs, Ys, masks = map(list, zip(*data_batch))

    Xs = torch.stack(Xs, dim=0)
    Ys = torch.stack(Ys, dim=0)
    masks = torch.stack(masks, dim=0)

    return Xs, Ys, masks


def set_random_seed(seed=0):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)  # 为CPU设置种子用于生成随机数
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)  # 为当前GPU设置随机种子


import os.path

import pandas as pd
from rdkit.Chem import AllChem
import numpy as np
from sklearn import preprocessing
def saltremover(i):
    l = i.split('.')
    d = {len(a):a for a in l }
    smile = d[max(d.keys())]
    return smile
def stand_smiles(smiles):
    try:
        smiles = AllChem.MolToSmiles(AllChem.MolFromSmiles(smiles))
    except:
        smiles = ''
    return smiles
def process(file,content,cpu=10):
    if content == 'cano':
        data = pd.read_csv(file)
        start =len(data)
        data['Smiles'] = data['Smiles'].apply(saltremover)
        data['Smiles'] = data['Smiles'].apply(stand_smiles)
        output = file.split('.csv')[0] + '_pro.csv'
        if os.path.exists(output):
            pass
        else:
            data.to_csv(output,index=False)
            if start-len(data)>0:
                print('we meet some smiles which cannot revert to cano_smiles and the number is',start-len(data))
    if content == 'descriptor':
        from padelpy import padeldescriptor
        output = file.split('.csv')[0] + '_pro.csv'
        data = pd.read_csv(output)
        data['activity'] = data['Smiles']
        data=data[['Smiles','activity']]
        smi = file.split('.csv')[0] +'.smi'
        des = file.split('.csv')[0] +'_23d.csv'
        if os.path.exists(des):
            pass
        else:
            data.to_csv(smi,index=False,sep='\t',header=None)
            padeldescriptor(mol_dir=smi, d_2d=True, d_3d=True, d_file=des,threads=50)
            # print('done 2d3d',end=' ')
    if content == 'pubchem':
        from padelpy import padeldescriptor
        output = file.split('.csv')[0] + '_pro.csv'
        data = pd.read_csv(output)
        data['activity'] = data['Smiles']
        data=data[['Smiles','activity']]
        smi = file.split('.csv')[0] +'.smi'
        des = file.split('.csv')[0] +'_pubchem.csv'
        if os.path.exists(des):
            pass
        else:
            data.to_csv(smi,index=False,sep='\t',header=None)
            padeldescriptor(mol_dir=smi, fingerprints=True, d_file=des,threads=30)
            # print('done punchem',end=' ')
    if content == 'adj_23d':
        des = file.split('.csv')[0] + '_23d.csv'
        name = file.split('.csv')[0] + '_23d_adj.csv'
        if os.path.exists(name):
            pass
        else:
            des = pd.read_csv(des).iloc[:,:1445]   # save pre 1444 features
            des = des.replace([np.inf,-np.inf],np.nan)  # trans inf to nan
            des = des.dropna(thresh=int(des.shape[1]/2),axis =0)  #drop that row all 0 how ='all'
            des = des.dropna(thresh=int(des.shape[0]/2),axis =1)
            des = des.fillna(0)  # fill nan by mean col
            min_max_scaler = preprocessing.MinMaxScaler()
            adj = min_max_scaler.fit_transform(des.drop(['Name'], axis=1))  # scaler
            adj = pd.DataFrame(adj,columns=list(des.columns)[1:])
            adj['Name']=des['Name']
            adj.to_csv(name,index=False)
            # print('done scaler',end =' ')
def start(file,des= None):
    content = ['cano']
    if '2d-3d' in des:
        content.extend(['descriptor','adj_23d'])
    if 'pubchem' in des:
        content.append('pubchem')
    for pro in content:
        process(file,pro)




import pandas as pd
import os
import warnings
warnings.filterwarnings('ignore')

def get_info(file_name=None,models=None,split=None, FP=None):
    dataset = file_name.split('/')[-2]
    path = os.path.join(os.path.join(file_name.split(dataset)[0],dataset),'result_save')
    des_model = [model for model in models if model in ['KNN','SVM','RF','XGB','DNN']]
    graph_model = [model for model in models if model not in ['KNN','SVM','RF','XGB','DNN']]
    info =pd.DataFrame()
    for model in des_model:
        model_dir = os.path.join(path,model)
        for des in FP:
            for sp in split:
                file = os.path.join(model_dir,'_'.join([sp,model,des,'best.csv']))
                df = pd.read_csv(file)
                data= df.groupby('type')['se','precision','auc_roc'].mean()
                data=data.reset_index()
                data['model']=model
                data['des'] = des
                data['split'] = sp
                info = pd.concat([info,data],ignore_index=True)
    for model in graph_model:
        model_dir = os.path.join(path,model)

        for sp in split:
            file = os.path.join(model_dir,'_'.join([sp,model,'best.csv']))
            df = pd.read_csv(file)
            data= df.groupby('type')['se','precision','auc_roc'].mean()
            data = data.reset_index()
            data['model']=model
            data['des'] = model
            data['split'] = sp
            info = pd.concat([info,data],ignore_index=True)
    info =info.sort_values('auc_roc',ascending=False)
    info_save = file_name.replace('.csv','_auc.csv')
    info.to_csv(info_save,index=False)

# def show_model(file_name=None,models=None,split=None, FP=None):
#     dataset = file_name.split('/')[-2]
#     model_path = os.path.join(os.path.join(file_name.split(dataset)[0], dataset), 'model_save')
#     para_path = os.path.join(os.path.join(file_name.split(dataset)[0], dataset), 'model_save')
#     des_model = [model for model in models if model in ['KNN', 'SVM', 'RF', 'XGB', 'DNN']]
#     graph_model = [model for model in models if model not in ['KNN', 'SVM', 'RF', 'XGB', 'DNN']]
#     for model in des_model:
#         if model =='DNN':
#             pass
#         else:
#             for sp in split:
#                 for des in FP:
#                     param_file = '_'.join([sp,model,des,'para.csv'])
def para(file_name):
    info_save = file_name.replace('.csv', '_auc.csv')
    data = pd.read_csv(info_save)

    data=data[data['type']=='te']
    data = data.sort_values('auc_roc', ascending=False)
    data['a_sord'] = [i for i in range(len(data))]
    data['f1-score'] = data.apply(lambda x:(2*(x['precision']*x['se'])/(x['precision']+x['se'])),axis=1)
    data=data.sort_values('f1-score',ascending=False)
    data['i_sord'] = [i for i in range(len(data))]
    data['value'] = data.apply(lambda x:x['a_sord']+x['i_sord'],axis=1)
    data = data.sort_values('value')
    data=data[[ 'auc_roc', 'model', 'des', 'split', 'f1-score']]
    print(data)


para('/data/jianping/bokey/OCAICM/dataset/chembl31_1663577628.6572962/chembl31_1663577628.6572962.csv')


from rdkit import Chem
import datetime
import dgl
import numpy as np
import random
import torch
import torch.nn.functional as F
import dgl
from dgl import model_zoo
from dgl.data.chem import one_hot_encoding
from dgl.data.utils import split_dataset
from sklearn.metrics import roc_auc_score
from dgl.data.chem import BaseAtomFeaturizer, BaseBondFeaturizer, ConcatFeaturizer, atom_type_one_hot, atom_degree_one_hot, \
    atom_implicit_valence_one_hot, atom_formal_charge, atom_num_radical_electrons, atom_hybridization_one_hot, \
    atom_is_aromatic, atom_total_num_H_one_hot, atom_total_degree_one_hot, atom_chiral_tag_one_hot, atom_mass, \
    bond_type_one_hot, bond_is_conjugated, bond_is_in_ring, bond_stereo_one_hot
from functools import partial
import torch.nn as nn
import dgl
import torch
from dgl.nn.pytorch.conv import RelGraphConv
from dgl.model_zoo.chem.classifiers import MLPBinaryClassifier
from dgl.nn.pytorch.glob import WeightAndSum
# from dgl import BatchedDGLGraph
from sklearn.metrics import roc_auc_score, confusion_matrix, precision_recall_curve, auc, \
    mean_absolute_error, r2_score, matthews_corrcoef


def chirality(atom):  # the chirality information defined in the AttentiveFP
    try:
        return one_hot_encoding(atom.GetProp('_CIPCode'), ['R', 'S']) + \
               [atom.HasProp('_ChiralityPossible')]
    except:
        return [False, False] + [atom.HasProp('_ChiralityPossible')]


AttentiveFPAtomFeaturizer = BaseAtomFeaturizer(
    featurizer_funcs={'h': ConcatFeaturizer([
        partial(atom_type_one_hot, allowable_set=[
            'B', 'C', 'N', 'O', 'F', 'Si', 'P', 'S', 'Cl', 'As', 'Se', 'Br', 'Te', 'I', 'At'],
                encode_unknown=True),
        partial(atom_degree_one_hot, allowable_set=list(range(6))),
        atom_formal_charge, atom_num_radical_electrons,
        partial(atom_hybridization_one_hot, encode_unknown=True),
        atom_is_aromatic,  # A placeholder for aromatic information,
        atom_total_num_H_one_hot, chirality
    ],
    )})

AttentiveFPBondFeaturizer = BaseBondFeaturizer(featurizer_funcs={'e': ConcatFeaturizer([bond_type_one_hot,
                                                                                        bond_is_conjugated,
                                                                                        bond_is_in_ring,
                                                                                        partial(bond_stereo_one_hot,
                                                                                                allowable_set=[Chem.rdchem.BondStereo.STEREONONE,
                                                                                                               Chem.rdchem.BondStereo.STEREOANY,
                                                                                                               Chem.rdchem.BondStereo.STEREOZ,
                                                                                                               Chem.rdchem.BondStereo.STEREOE], encode_unknown=True)])})


# AtomFeaturizer.feat_size('h') = 85
class MyAtomFeaturizer(BaseAtomFeaturizer):
    def __init__(self, atom_data_field='h'):
        super(MyAtomFeaturizer, self).__init__(
            featurizer_funcs={atom_data_field: ConcatFeaturizer(
                [atom_type_one_hot,
                 atom_degree_one_hot,
                 atom_implicit_valence_one_hot,
                 atom_formal_charge,
                 atom_num_radical_electrons,
                 atom_hybridization_one_hot,
                 atom_is_aromatic,
                 atom_total_num_H_one_hot,
                 atom_total_degree_one_hot,
                 atom_chiral_tag_one_hot,
                 atom_mass]
            )})


def set_random_seed(seed=0):
    """Set random seed.

    Parameters
    ----------
    seed : int
        Random seed to use
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)  # 为CPU设置种子用于生成随机数
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)  # 为当前GPU设置随机种子


# class Meter(object):
#     """Track and summarize model performance on a dataset for
#     (multi-label) binary classification."""
#     def __init__(self):
#         self.mask = []
#         self.y_pred = []
#         self.y_true = []
#
#     def update(self, y_pred, y_true, mask):
#         """Update for the result of an iteration
#
#         Parameters
#         ----------
#         y_pred : float32 tensor
#             Predicted molecule labels with shape (B, T),
#             B for batch size and T for the number of tasks
#         y_true : float32 tensor
#             Ground truth molecule labels with shape (B, T)
#         mask : float32 tensor
#             Mask for indicating the existence of ground
#             truth labels with shape (B, T)
#         """
#         self.y_pred.append(y_pred.detach().cpu())
#         self.y_true.append(y_true.detach().cpu())
#         self.mask.append(mask.detach().cpu())
#
#     def roc_auc_score(self):
#         """Compute roc-auc score for each task.
#
#         Returns
#         -------
#         list of float
#             roc-auc score for all tasks
#         """
#         mask = torch.cat(self.mask, dim=0)
#         y_pred = torch.cat(self.y_pred, dim=0)
#         y_true = torch.cat(self.y_true, dim=0)
#         # Todo: support categorical classes
#         # This assumes binary case only
#         y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率
#         n_tasks = y_true.shape[1]
#         scores = []
#         for task in range(n_tasks):
#             task_w = mask[:, task]
#             task_y_true = y_true[:, task][task_w != 0].numpy()
#             task_y_pred = y_pred[:, task][task_w != 0].numpy()
#             scores.append(roc_auc_score(task_y_true, task_y_pred))
#         return scores
#
#     def l1_loss(self, reduction):
#         """Compute l1 loss for each task.
#
#         Returns
#         -------
#         list of float
#             l1 loss for all tasks
#         reduction : str
#             * 'mean': average the metric over all labeled data points for each task
#             * 'sum': sum the metric over all labeled data points for each task
#         """
#         mask = torch.cat(self.mask, dim=0)
#         y_pred = torch.cat(self.y_pred, dim=0)
#         y_true = torch.cat(self.y_true, dim=0)
#         n_tasks = y_true.shape[1]
#         scores = []
#         for task in range(n_tasks):
#             task_w = mask[:, task]
#             task_y_true = y_true[:, task][task_w != 0]
#             task_y_pred = y_pred[:, task][task_w != 0]
#             scores.append(F.l1_loss(task_y_true, task_y_pred, reduction=reduction).item())
#         return scores
#
#     def rmse(self):
#         """Compute RMSE for each task.
#
#         Returns
#         -------
#         list of float
#             rmse for all tasks
#         """
#         mask = torch.cat(self.mask, dim=0)
#         y_pred = torch.cat(self.y_pred, dim=0)
#         y_true = torch.cat(self.y_true, dim=0)
#         n_data, n_tasks = y_true.shape
#         scores = []
#         for task in range(n_tasks):
#             task_w = mask[:, task]
#             task_y_true = y_true[:, task][task_w != 0]
#             task_y_pred = y_pred[:, task][task_w != 0]
#             scores.append(np.sqrt(F.mse_loss(task_y_pred, task_y_true).cpu().item()))
#         return scores
#
#     def compute_metric(self, metric_name, reduction='mean'):
#         """Compute metric for each task.
#
#         Parameters
#         ----------
#         metric_name : str
#             Name for the metric to compute.
#         reduction : str
#             Only comes into effect when the metric_name is l1_loss.
#             * 'mean': average the metric over all labeled data points for each task
#             * 'sum': sum the metric over all labeled data points for each task
#
#         Returns
#         -------
#         list of float
#             Metric value for each task
#         """
#         assert metric_name in ['roc_auc', 'l1', 'rmse'], \
#             'Expect metric name to be "roc_auc", "l1" or "rmse", got {}'.format(metric_name)  # assert（断言）用于判断一个表达式，在表达式条件为 false 的时候触发异常
#         assert reduction in ['mean', 'sum']
#         if metric_name == 'roc_auc':
#             return self.roc_auc_score()
#         if metric_name == 'l1':
#             return self.l1_loss(reduction)
#         if metric_name == 'rmse':
#             return self.rmse()
class Meter(object):
    """Track and summarize model performance on a dataset for
    (multi-label) binary classification."""
    def __init__(self):
        self.mask = []
        self.y_pred = []
        self.y_true = []

    def update(self, y_pred, y_true, mask):
        """Update for the result of an iteration

        Parameters
        ----------
        y_pred : float32 tensor
            Predicted molecule labels with shape (B, T),
            B for batch size and T for the number of tasks
        y_true : float32 tensor
            Ground truth molecule labels with shape (B, T)
        mask : float32 tensor
            Mask for indicating the existence of ground
            truth labels with shape (B, T)
        """

        self.y_pred.append(y_pred.detach().cpu())
        self.y_true.append(y_true.detach().cpu())
        self.mask.append(mask.detach().cpu())

    def roc_precision_recall_score(self):
        """Compute AUC_PRC for each task.
        Returns
        -------
        list of float
            rmse for all tasks
        """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_pred = torch.sigmoid(y_pred)
        y_true = torch.cat(self.y_true, dim=0)
        n_data, n_tasks = y_true.shape
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()
            precision, recall, _thresholds = precision_recall_curve(task_y_true, task_y_pred, pos_label=1)
            scores.append(auc(recall, precision))
        return scores

    def roc_auc_score(self):
        """Compute roc-auc score for each task.

        Returns
        -------
        list of float
            roc-auc score for all tasks
        """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        # Todo: support categorical classes
        # This assumes binary case only
        y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率

        n_tasks = y_true.shape[1]
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()

            scores.append(roc_auc_score(task_y_true, task_y_pred))
        return scores

    def l1_loss(self, reduction):
        """Compute l1 loss for each task.

        Returns
        -------
        list of float
            l1 loss for all tasks
        reduction : str
            * 'mean': average the metric over all labeled data points for each task
            * 'sum': sum the metric over all labeled data points for each task
        """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        n_tasks = y_true.shape[1]
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0]
            task_y_pred = y_pred[:, task][task_w != 0]
            scores.append(F.l1_loss(task_y_true, task_y_pred, reduction=reduction).item())
        return scores

    def rmse(self):
        """Compute RMSE for each task.

        Returns
        -------
        list of float
            rmse for all tasks
        """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        n_data, n_tasks = y_true.shape
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0]
            task_y_pred = y_pred[:, task][task_w != 0]
            scores.append(np.sqrt(F.mse_loss(task_y_pred, task_y_true).cpu().item()))
        return scores

    def mae(self):
        """Compute mae for each task.

        Returns
        -------
        list of float
            mae for all tasks
        """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        n_data, n_tasks = y_true.shape
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0]
            task_y_pred = y_pred[:, task][task_w != 0]
            scores.append(mean_absolute_error(task_y_true, task_y_pred))
        return scores

    def se(self):
        """Compute se score for each task.

                Returns
                -------
                list of float
                    se score for all tasks
                """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        # Todo: support categorical classes
        # This assumes binary case only
        y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率
        n_tasks = y_true.shape[1]
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()

            task_y_pred = [1 if i >= 0.5 else 0 for i in task_y_pred]
            c_mat = confusion_matrix(task_y_true, task_y_pred)

            tn, fp, fn, tp = list(c_mat.flatten())
            se = tp / (tp + fn)
            scores.append(se)

        return scores
    def precision(self):
        """Compute precision score for each task.

                Returns
                -------
                list of float
                    precision score for all tasks
                """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        # Todo: support categorical classes
        # This assumes binary case only
        y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率
        n_tasks = y_true.shape[1]
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()

            task_y_pred = [1 if i >=0.5 else 0 for i in task_y_pred]
            c_mat = confusion_matrix(task_y_true, task_y_pred)

            tn, fp, fn, tp = list(c_mat.flatten())
            precision = tp / (tp + fp)
            scores.append(precision)

        return scores
    def sp(self):
        """Compute sp score for each task.

                Returns
                -------
                list of float
                    sp score for all tasks
                """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        # Todo: support categorical classes
        # This assumes binary case only
        y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率
        n_tasks = y_true.shape[1]
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()

            task_y_pred = [1 if i >=0.5 else 0 for i in task_y_pred]
            c_mat = confusion_matrix(task_y_true, task_y_pred)

            tn, fp, fn, tp = list(c_mat.flatten())

            sp = tn / (tn + fp)

            scores.append(sp)

        return scores
    def acc(self):
        """Compute acc score for each task.

                Returns
                -------
                list of float
                    acc score for all tasks
                """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        # Todo: support categorical classes
        # This assumes binary case only
        y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率
        n_tasks = y_true.shape[1]
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()

            task_y_pred = [1 if i >=0.5 else 0 for i in task_y_pred]
            c_mat = confusion_matrix(task_y_true, task_y_pred)

            tn, fp, fn, tp = list(c_mat.flatten())

            acc = (tp + tn) / (tn + fp + fn + tp)

            scores.append(acc)

        return scores
    def mcc(self):
        """Compute mcc score for each task.

                Returns
                -------
                list of float
                    mcc score for all tasks
                """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        # Todo: support categorical classes
        # This assumes binary case only
        y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率
        n_tasks = y_true.shape[1]
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0].numpy()
            task_y_pred = y_pred[:, task][task_w != 0].numpy()
            task_y_pred = [1 if i >=0.5 else 0 for i in task_y_pred]
            c_mat = confusion_matrix(task_y_true, task_y_pred)
            tn, fp, fn, tp = list(c_mat.flatten())
            mcc = (tp * tn - fp * fn) / np.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn) + 1e-8)
            scores.append(mcc)

        return scores
    def r2(self):
        """Compute r2 score for each task.

        Returns
        -------
        list of float
            r2 score for all tasks
        """
        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)
        y_true = torch.cat(self.y_true, dim=0)
        n_data, n_tasks = y_true.shape
        scores = []
        for task in range(n_tasks):
            task_w = mask[:, task]
            task_y_true = y_true[:, task][task_w != 0]
            task_y_pred = y_pred[:, task][task_w != 0]
            scores.append(r2_score(task_y_true, task_y_pred))
        return scores

    def pred(self):

        mask = torch.cat(self.mask, dim=0)
        y_pred = torch.cat(self.y_pred, dim=0)

        y_pred = torch.sigmoid(y_pred)  # 求得为正例的概率

        return y_pred
    def compute_metric(self, metric_name, reduction='mean'):
        """Compute metric for each task.

        Parameters
        ----------
        metric_name : str
            Name for the metric to compute.
        reduction : str
            Only comes into effect when the metric_name is l1_loss.
            * 'mean': average the metric over all labeled data points for each task
            * 'sum': sum the metric over all labeled data points for each task

        Returns
        -------
        list of float
            Metric value for each task
        """
        assert metric_name in ['roc_auc', 'l1', 'rmse', 'prc_auc', 'mae', 'r2','se','sp','acc','mcc','pred','precision'], \
            'Expect metric name to be "roc_auc", "l1", "rmse", "prc_auc", "mae", "r2","pred" got {}'.format(metric_name)  # assert（断言）用于判断一个表达式，在表达式条件为 false 的时候触发异常
        assert reduction in ['mean', 'sum']
        if metric_name == 'roc_auc':
            return self.roc_auc_score()
        if metric_name == 'l1':
            return self.l1_loss(reduction)
        if metric_name == 'rmse':
            return self.rmse()
        if metric_name == 'prc_auc':
            return self.roc_precision_recall_score()
        if metric_name == 'se':
            return self.se()

        if metric_name == 'precision':
            return self.precision()
        if metric_name == 'sp':
            return self.sp()
        if metric_name == 'acc':
            return self.acc()
        if metric_name == 'mcc':
            return self.mcc()
        if metric_name == 'mae':
            return self.mae()
        if metric_name == 'r2':
            return self.r2()
        if metric_name == 'mcc':
            return self.mcc()
        if metric_name == 'pred':
            return self.pred()

class EarlyStopping(object):
    """Early stop performing

    Parameters
    ----------
    mode : str
        * 'higher': Higher metric suggests a better model
        * 'lower': Lower metric suggests a better model
    patience : int
        Number of epochs to wait before early stop
        if the metric stops getting improved
    filename : str or None
        Filename for storing the model checkpoint
    """
    def __init__(self, mode='higher', patience=10, filename=None):  # 可以加tolerance
        if filename is None:
            dt = datetime.datetime.now()
            filename = '{}_early_stop_{}_{:02d}-{:02d}-{:02d}.pth'.format(
                dt.date(), dt.hour, dt.minute, dt.second)

        assert mode in ['higher', 'lower']
        self.mode = mode
        if self.mode == 'higher':
            self._check = self._check_higher  # 私有变量（private），只有内部可以访问，外部不能访问
        else:
            self._check = self._check_lower

        self.patience = patience
        self.counter = 0
        self.filename = filename
        self.best_score = None
        self.early_stop = False

    def _check_higher(self, score, prev_best_score):
        return (score > prev_best_score)

    def _check_lower(self, score, prev_best_score):
        return (score < prev_best_score)

    def step(self, score, model):
        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(model)
        elif self._check(score, self.best_score):  # 当前模型如果是更优模型，则保存当前模型
            self.best_score = score
            self.save_checkpoint(model)
            self.counter = 0
        else:
            self.counter += 1
            # print(
            #     f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
        return self.early_stop

    def save_checkpoint(self, model):
        '''Saves model when the metric on the validation set gets improved.'''
        torch.save({'model_state_dict': model.state_dict()}, self.filename)

    def load_checkpoint(self, model):
        '''Load model saved with early stopping.'''
        model.load_state_dict(torch.load(self.filename)['model_state_dict'])
# class EarlyStopping(object):
#     def __init__(self, args, mode='higher', patience=10, filename=None):
#         if filename is None:
#             dt = datetime.datetime.now()
#             filename = './saved_model/' + '{}_{}_early_stop_{}_{:02d}-{:02d}-{:02d}.pth'.format(args['task'],args['model'], dt.date(), dt.hour, dt.minute,
#                                                                                                 dt.second)
#
#         assert mode in ['higher', 'lower']
#         self.mode = mode
#         if self.mode == 'higher':
#             self._check = self._check_higher  # 私有变量（private），只有内部可以访问，外部不能访问
#         else:
#             self._check = self._check_lower
#
#         self.patience = patience
#         self.counter = 0
#         self.filename = filename
#         self.best_score = None
#         self.early_stop = False
#
#     def _check_higher(self, score, prev_best_score):
#         return (score > prev_best_score)
#
#     def _check_lower(self, score, prev_best_score):
#         return (score < prev_best_score)
#
#     def step(self, score, model):
#         if self.best_score is None:
#             self.best_score = score
#             self.save_checkpoint(model)
#         elif self._check(score, self.best_score):
#             self.best_score = score
#             self.save_checkpoint(model)
#             self.counter = 0
#         else:
#             self.counter += 1
#             # print(
#             # f'EarlyStopping counter: {self.counter} out of {self.patience}')
#             if self.counter >= self.patience:
#                 self.early_stop = True
#         return self.early_stop
#
#     def save_checkpoint(self, model):
#         '''Saves model when the metric on the validation set gets improved.'''
#         torch.save({'model_state_dict': model.state_dict()}, self.filename)
#
#     def load_checkpoint(self, model):
#         '''Load model saved with early stopping.'''
#         model.load_state_dict(torch.load(self.filename)['model_state_dict'])

def collate_molgraphs(data):
    assert len(data[0]) in [3, 4], \
        'Expect the tuple to be of length 3 or 4, got {:d}'.format(len(data[0]))
    if len(data[0]) == 3:
        smiles, graphs, labels = map(list, zip(*data))
        masks = None
    else:
        smiles, graphs, labels, masks = map(list, zip(*data))

    bg = dgl.batch(graphs)
    bg.set_n_initializer(dgl.init.zero_initializer)
    bg.set_e_initializer(dgl.init.zero_initializer)
    labels = torch.stack(labels, dim=0)

    if masks is None:
        masks = torch.ones(labels.shape)
    else:
        masks = torch.stack(masks, dim=0)
    return smiles, bg, labels, masks





import pandas as pd
import multiprocessing as mp
from padelpy import from_smiles

cpu= 20
com = len(pd.read_csv('/data/jianping/bokey/OCAICM/dataset/chembl31/error.txt'))
num = int(com/cpu) +1
df = pd.read_csv('/data/jianping/bokey/OCAICM/dataset/chembl31/error.txt')
def convert(data,num,count):

    start ,end= count*num,(count+1)*num-1
    end  = end if end<len(data) else len(data)
    data = data.iloc[start:end,:]

    for smiles in data['Smiles'].tolist():
        try:
            info = from_smiles(smiles,descriptors=True,timeout=100)
            f = open('/data/jianping/bokey/OCAICM/dataset/chembl31/target.txt','a')
            info =list(info.items())
            info = [value[1] for value in info]
            f.write('{},{}\n'.format(smiles,''.join(info)))
            f.close()
        except:
            e = open('/data/jianping/bokey/OCAICM/dataset/chembl31/error.txt', 'a')
            e.write('{}\n'.format(smiles))
            e.close()
p = mp.Pool(processes = cpu)
for i in range(cpu):
    result = p.apply_async(convert, args=(df, num,i))
p.close()
p.join()


import warnings
import numpy as np
import pandas as pd
import argparse
import random as rn
import multiprocessing as mp
import time
import os
import sys
from bokeys.XGB_model import tvt_xgb
from bokeys.SVM_model import tvt_svm
from bokeys.RF_model import tvt_rf
from bokeys.KNN_model import tvt_knn
from bokeys.DNN_model import tvt_dnn
from bokeys.graph_model import tvt_dl
import itertools
np.set_printoptions(threshold=sys.maxsize)
#Random seeding
os.environ['PYTHONHASHSEED']='0'
np.random.seed(123)
rn.seed(123)
warnings.filterwarnings('ignore')
def file_merge(file_df,models):
    filelist={}
    dataset = file_df.split('/')[-1].replace('.csv','')
    dir = file_df.split(dataset)[-1]
    df =pd.read_csv(file_df)
    cols = list(df.columns)
    cols.remove('Smiles')
    path = '{}{}/result_save'.format(dir,dataset)
    for task in cols:
        for model in models:
            for file in os.listdir(os.path.join(os.path.join(path,task),model)):
                if 'para' in file:
                    continue
                filelist.setdefault(file,[]).append(os.path.join(os.path.join(os.path.join(os.path.join(path,task),model)),file))
    for rtype in filelist.keys():
        rlist = filelist[rtype]
        mer = pd.DataFrame()
        for file in rlist:
            df = pd.read_csv(file)

            mer = pd.concat([mer, df], ignore_index=True)
        mer = mer.groupby(['seed', 'FP_type', 'split_type', 'type'])[
            ['se', 'sp', 'acc', 'mcc', 'precision','auc_prc', 'auc_roc']].mean()
        mer = mer.reset_index()
        save_path = os.path.join(path, rtype.split('_')[1])

        if not os.path.exists(save_path):
            os.makedirs(save_path)

        abspath = os.path.join(save_path, rtype)
        mer.to_csv(abspath, index=False)
    pass

def model_set(X,Y,split_type='random',FP_type='ECFP4',model_type = 'SVM',model_dir=False,file_name=None,task_type='cla',difftasks=None):
    if model_type =='SVM':
        para,best = tvt_svm(X, Y, split_type=split_type, FP_type=FP_type,model_dir=model_dir)
    elif model_type =='RF':
        para,best = tvt_rf(X, Y, split_type=split_type, FP_type=FP_type,model_dir=model_dir)
    elif model_type =='KNN':
        para,best = tvt_knn(X, Y, split_type=split_type, FP_type=FP_type,model_dir=model_dir)
    elif model_type == 'XGB':
        para, best = tvt_xgb(X, Y, split_type=split_type, FP_type=FP_type,model_dir=model_dir)
    elif model_type == 'DNN':
        para, best = tvt_dnn(X, Y, split_type=split_type, FP_type=FP_type,model_dir=model_dir,file_name=file_name,difftasks=difftasks)
    elif model_type == 'gat':
        para, best = tvt_dl(X, Y, split_type=split_type,file_name=file_name,model_name='gat',model_dir=model_dir,difftasks=difftasks,task_type=task_type,device ='cpu')
    elif model_type == 'attentivefp':
        para, best = tvt_dl(X, Y, split_type=split_type,file_name=file_name,model_name='attentivefp',model_dir=model_dir,difftasks=difftasks,task_type=task_type,device ='cpu')
    elif model_type == 'gcn':
        para, best = tvt_dl(X, Y, split_type=split_type,file_name=file_name,model_name='gcn',model_dir=model_dir,difftasks=difftasks,task_type=task_type,device ='cpu')
    else:
        para, best = tvt_dl(X, Y, split_type=split_type,file_name=file_name,model_name='mpnn',model_dir=model_dir,difftasks=difftasks,task_type=task_type,device ='cpu')
    return para,best

def pair_param(file_name,data,split_type,model_type,FP_type,difftasks):

    if model_type in ['SVM','KNN','RF','XGB']:
        if len(difftasks) == 1:

            X = data.Smiles

            Y = data[difftasks[0]]
            model_path = 'model_save/'+model_type
            model_dir = file_name.replace(file_name.split('/')[-1],model_path)

            if not os.path.exists(model_dir):
                os.makedirs(model_dir)
            result_path = 'result_save/' + model_type
            result_dir = file_name.replace(file_name.split('/')[-1], result_path)
            if not os.path.exists(result_dir):
                os.makedirs(result_dir)
            para_name, best_name = os.path.join(result_dir,
                                                    '_'.join([split_type, model_type, FP_type, 'para.csv'])), os.path.join(
                    result_dir, '_'.join([split_type, model_type, FP_type, 'best.csv']))
            if os.path.exists(para_name):
                pass
            else:
                para,best = model_set(X,Y, split_type=split_type,model_type=model_type,model_dir=model_dir,file_name=file_name,FP_type=FP_type)#
                para.to_csv(para_name,index=False)
                best.to_csv(best_name,index=False)
        else:
            for task in difftasks:
                rm_index = data[task][data[task].isnull()].index
                data.drop(index=rm_index, inplace=True)
                Y = data[task]
                X = data.Smiles
                model_path = 'model_save/'+task +'/'+ model_type
                model_dir = file_name.replace(file_name.split('/')[-1], model_path)
                if not os.path.exists(model_dir):
                    os.makedirs(model_dir)
                result_path = 'result_save/' + task + '/' + model_type
                result_dir = file_name.replace(file_name.split('/')[-1], result_path)
                if not os.path.exists(result_dir):
                    os.makedirs(result_dir)
                para_name, best_name = os.path.join(result_dir,
                                                        '_'.join([split_type, model_type, FP_type, 'para.csv'])), os.path.join(
                    result_dir, '_'.join([split_type, model_type, FP_type, 'best.csv']))
                if os.path.exists(para_name):
                    pass
                else:
                    para, best = model_set(X,Y, split_type=split_type,model_type=model_type,model_dir=model_dir,file_name=file_name,FP_type=FP_type)


                    para.to_csv(para_name, index=False)
                    best.to_csv(best_name, index=False)
    elif model_type in ['DNN']:
        X = data.Smiles
        Y = data[difftasks[0]]
        model_path = 'model_save/' + model_type
        model_dir = file_name.replace(file_name.split('/')[-1], model_path)

        if not os.path.exists(model_dir):
            os.makedirs(model_dir)
        result_path = 'result_save/' + model_type
        result_dir = file_name.replace(file_name.split('/')[-1], result_path)
        if not os.path.exists(result_dir):
            os.makedirs(result_dir)
        para_name, best_name = os.path.join(result_dir,
                                            '_'.join([split_type, model_type, FP_type, 'para.csv'])), os.path.join(
            result_dir, '_'.join([split_type, model_type, FP_type, 'best.csv']))
        if os.path.exists(para_name):
            print(para_name, 'has done')
            pass
        else:
            para, best = model_set(X, Y, split_type=split_type, model_type=model_type, model_dir=model_dir,
                                   file_name=file_name, FP_type=FP_type, difftasks=difftasks)  #
            para.to_csv(para_name, index=False)
            best.to_csv(best_name, index=False)
    else:
        X = data.Smiles
        Y = data[difftasks[0]]
        model_path = 'model_save/' + model_type
        model_dir = file_name.replace(file_name.split('/')[-1], model_path)
        result_path = 'result_save/' + model_type
        result_dir = file_name.replace(file_name.split('/')[-1], result_path)
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)
        if not os.path.exists(result_dir):
            os.makedirs(result_dir)
        para_name, best_name = os.path.join(result_dir, '_'.join([split_type, model_type, 'para.csv'])), os.path.join(
            result_dir, '_'.join([split_type, model_type, 'best.csv']))
        if os.path.exists(para_name):
            print(para_name, "exists files")
            pass
        else:
            para, best = model_set(X, Y, split_type=split_type, model_type=model_type, model_dir=model_dir,
                                   file_name=file_name, difftasks=difftasks)
            para.to_csv(para_name, index=False)
            best.to_csv(best_name, index=False)

        return
    return

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--file', required=True, help='we must give this para')
    parser.add_argument('--split', default=['scaffold'], nargs='*',choices = ['random','scaffold','cluster'])
    parser.add_argument('--FP', default=['MACCS'], nargs='*',choices = ['ECFP4','MACCS','2d-3d','pubchem'])
    parser.add_argument('--model', default=['SVM'], nargs='*',choices = ['DNN','KNN','SVM','RF','XGB'])
    parser.add_argument('--threads', default=10,type=int)
    parser.add_argument('--mpl', default=False, type=str)
    parser.add_argument('--device', default='cpu', choices=['cpu', 'gpu'])
    args = parser.parse_args()
    return args

def main(file,split,FP,models,cpus,mpl,device):
    ml_models = [model for model in models if model in ['KNN','SVM','RF','XGB']]
    ml = {'split':split,'FP':FP,'model':ml_models}
    ml = pd.DataFrame(list(itertools.product(*ml.values())),columns=ml.keys())
    dl_models = [model for model in models if model in ['gcn','gat','mpnn','attentivefp']]
    dl = {'split':split,'model':dl_models}
    dl = pd.DataFrame(list(itertools.product(*dl.values())),columns=dl.keys())
    data = pd.read_csv(file, error_bad_lines=False)
    difftasks = list(data.columns)
    difftasks.remove('Smiles')

    if mpl == False:
        for i in range(len(ml.index.values)):
            start_time = time.time()
            split_type = ml.iloc[i].split
            FP_type = ml.iloc[i].FP
            model_type = ml.iloc[i].model
            a = pair_param(file, data, split_type, model_type, FP_type,difftasks)
            use_time = round(time.time() - start_time,5)
            if use_time >=1:
                local_time = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())
                record_save = file.replace(file.split('/')[-1],'record_every_model.csv')
                f = open(record_save, 'a+')
                f.write(",".join([file, split_type, model_type, str(use_time),local_time,FP_type,'\n']))
                f.close()
    else:
        p = mp.Pool(processes = cpus)
        for i in range(len(ml.index.values)):
            start_time = time.time()
            split_type = ml.iloc[i].split
            FP_type = ml.iloc[i].FP
            model_type = ml.iloc[i].model
            result = p.apply_async(pair_param,args=(file,data,split_type,model_type,FP_type,difftasks))
            use_time = int(time.time() - start_time)
            if use_time >= 1:
                local_time = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())
                record_save = file.replace(file.split('/')[-1],'record_every_model.csv')
                f = open(record_save, 'a+')
                f.write(",".join([file, split_type, model_type, str(use_time),local_time,FP_type,'\n']))
                f.close()
        p.close()
        p.join()
    if 'DNN' in models:
        dnn = {'split':split,'FP':FP,'model':['DNN']}
        dnn = pd.DataFrame(list(itertools.product(*dnn.values())), columns=dnn.keys())
        for i in range(len(dnn.index.values)):
            start_time = time.time()
            split_type = dnn.iloc[i].split
            FP_type = dnn.iloc[i].FP
            model_type = dnn.iloc[i].model
            a = pair_param(file, data, split_type, model_type, FP_type, difftasks)
            use_time = round(time.time() - start_time, 5)
            local_time = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())
            if use_time >=1:
                record_save = file.replace(file.split('/')[-1],'record_every_model.csv')
                f = open(record_save, 'a+')
                f.write(",".join([file, split_type, model_type, str(use_time), local_time, FP_type, '\n']))
                f.close()
    for i in range(len(dl.index.values)):
        split_type = dl.iloc[i].split
        model_type = dl.iloc[i].model
        start_time = time.time()
        a = pair_param(file, data, split_type, model_type, device, difftasks)
        use_time = int(time.time() - start_time)
        if use_time >= 1:
            local_time = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())
            record_save = file.replace(file.split('/')[-1],'record_every_model.csv')
            f = open(record_save, 'a+')
            f.write(",".join([file, split_type, model_type, str(use_time), local_time, device, "\n"]))
            f.close()

    if len(difftasks) > 1:
        if len(set((['KNN', 'SVM', 'RF', 'XGB'])) | set(models)) > 0:
            need = list(set((['KNN', 'SVM', 'RF', 'XGB'])) & set(models))
            file_merge(file,need)


if __name__ == '__main__':
    args = parse_args()
    file = args.file
    split = args.split
    FP = args.FP
    model = args.model
    cpus=args.threads
    mpl = args.mpl
    device = args.device
    main(file,split,FP,model,cpus,mpl,device)


import time
import warnings
from bokeys.splitdater import split_dataset
from bokeys.feature_create import create_des
from hyperopt import fmin, tpe, hp, STATUS_OK, Trials
from sklearn.metrics import roc_auc_score, confusion_matrix, precision_recall_curve, auc, mean_squared_error, \
    r2_score, mean_absolute_error
from sklearn import svm
import pandas as pd
import numpy as np
import joblib
from sklearn import preprocessing
import os
def all_one_zeros(data):
    if (len(np.unique(data)) == 2):
        flag = False
    else:
        flag = True
    return flag


start = time.time()
warnings.filterwarnings("ignore")

# the metrics for classification
def statistical(y_true, y_pred, y_pro):
    c_mat = confusion_matrix(y_true, y_pred)
    tn, fp, fn, tp = list(c_mat.flatten())
    se = tp / (tp + fn)
    sp = tn / (tn + fp)
    precision = tp / (tp + fp)
    acc = (tp + tn) / (tn + fp + fn + tp)
    mcc = (tp * tn - fp * fn) / np.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn) + 1e-8)
    auc_prc = auc(precision_recall_curve(y_true, y_pro, pos_label=1)[1],
                  precision_recall_curve(y_true, y_pro, pos_label=1)[0])
    auc_roc = roc_auc_score(y_true, y_pro)
    return precision, se, sp, acc, mcc, auc_prc, auc_roc
def best_model_runing(seed,best_results,X,Y,split_type='random',FP_type='ECFP4',task_type = 'cla',model_dir=False):
    pd_res = []
    if task_type == 'cla':
        while True:

            data_tr_x, data_va_x, data_te_x, data_tr_y, data_va_y, data_te_y = split_dataset(X, Y,
                                                                                                 split_type=split_type,
                                                                                                 valid_need=True,
                                                                                                 random_state=seed)
            data_tr_x, data_tr_y = create_des(data_tr_x, data_tr_y, FP_type=FP_type,model_dir=model_dir)
            data_va_x, data_va_y = create_des(data_va_x, data_va_y, FP_type=FP_type,model_dir=model_dir)
            data_te_x, data_te_y = create_des(data_te_x, data_te_y, FP_type=FP_type,model_dir=model_dir)

            if (all_one_zeros(data_tr_y) or all_one_zeros(data_va_y) or all_one_zeros(data_te_y)):
                # print(
                #     '\ninvalid random seed {} due to one class presented in the {} splitted sets...'.format(seed,
                #                                                                                             split_type))
                # print('Changing to another random seed...\n')
                seed = np.random.randint(50, 999999)
            else:

                break
    else :
        data_tr_x, data_va_x, data_te_x, data_tr_y, data_va_y, data_te_y = split_dataset(X, Y, split_type=split_type,
                                                                                         valid_need=True,
                                                                                         random_state=seed)
        data_tr_x, data_tr_y = create_des(data_tr_x, data_tr_y, FP_type=FP_type)
        data_va_x, data_va_y = create_des(data_va_x, data_va_y, FP_type=FP_type)
        data_te_x, data_te_y = create_des(data_te_x, data_te_y, FP_type=FP_type)
    num_fea = data_tr_x.shape[1]
    model = svm.SVC(C=best_results['C'],
                    gamma=best_results['gamma'], kernel='rbf', random_state=1,
                    probability=True, class_weight='balanced',
                    cache_size=2000, max_iter=10000) \
        if task_type == 'cla' else svm.SVR(
        C=best_results['C'],
        gamma=best_results['gamma'],
        kernel='rbf', cache_size=2000, max_iter=10000)

    model.fit(data_tr_x, data_tr_y)
    num_of_compounds = X.shape[0]
    if task_type == 'cla':
        tr_pred = model.predict_proba(data_tr_x)
        tr_results = [seed, FP_type, split_type, 'tr', num_of_compounds]
        tr_results.extend(statistical(data_tr_y, np.argmax(tr_pred, axis=1), tr_pred[:, 1]))
        pd_res.append(tr_results)
        # validation set
        va_pred = model.predict_proba(data_va_x)
        va_results = [seed, FP_type, split_type, 'va', num_of_compounds]
        va_results.extend(statistical(data_va_y, np.argmax(va_pred, axis=1), va_pred[:, 1]))
        pd_res.append(va_results)
        # test set
        te_pred = model.predict_proba(data_te_x)
        te_results = [seed, FP_type, split_type, 'te', num_of_compounds]
        te_results.extend(statistical(data_te_y, np.argmax(te_pred, axis=1), te_pred[:, 1]))
        pd_res.append(te_results)
    else:
        # training set
        tr_pred = model.predict(data_tr_x)
        tr_results = [seed, FP_type, split_type, 'tr',  num_of_compounds,
                      np.sqrt(mean_squared_error(data_tr_y, tr_pred)), r2_score(data_tr_y, tr_pred),
                      mean_absolute_error(data_tr_y, tr_pred)]
        pd_res.append(tr_results)
        # validation set
        va_pred = model.predict(data_va_x)
        va_results = [seed, FP_type, split_type, 'va',  num_of_compounds,
                      np.sqrt(mean_squared_error(data_va_y, va_pred)), r2_score(data_va_y, va_pred),
                      mean_absolute_error(data_va_y, va_pred)]
        pd_res.append(va_results)
        # test set
        te_pred = model.predict(data_te_x)
        te_results = [seed, FP_type, split_type, 'te',  num_of_compounds,
                      np.sqrt(mean_squared_error(data_te_y, te_pred)), r2_score(data_te_y, te_pred),
                      mean_absolute_error(data_te_y, te_pred)]
        pd_res.append(te_results)
    return pd_res
def tvt_svm(X,Y,split_type='random',FP_type='ECFP4',task_type = 'cla',model_dir=False):

    data_tr_x, data_va_x, data_te_x, data_tr_y, data_va_y, data_te_y = split_dataset(X, Y, split_type=split_type,
                                                                                         valid_need=True,random_state=40)
    data_tr_x, data_tr_y = create_des(data_tr_x, data_tr_y, FP_type=FP_type,model_dir=model_dir)
    data_va_x, data_va_y = create_des(data_va_x, data_va_y, FP_type=FP_type,model_dir=model_dir)
    data_te_x, data_te_y = create_des(data_te_x, data_te_y, FP_type=FP_type,model_dir=model_dir)


    pd_res = []
    OPT_ITERS = 50
    space_ = {'C': hp.uniform('C', 0.1, 100), 'gamma': hp.uniform('gamma', 0, 0.2)}

    def hyper_opt(args):
        model = svm.SVC(**args, kernel='rbf', random_state=1, probability=True, class_weight='balanced',
                        cache_size=2000, max_iter=10000) if task_type == 'cla' else \
            svm.SVR(**args, kernel='rbf', cache_size=2000, max_iter=10000)

        model.fit(data_tr_x, data_tr_y)
        val_preds = model.predict_proba(data_va_x) if task_type == 'cla' else model.predict(data_va_x)
        loss = 1 - roc_auc_score(data_va_y, val_preds[:, 1]) if task_type == 'cla' else np.sqrt(
            mean_squared_error(data_va_y, val_preds))
        return {'loss': loss, 'status': STATUS_OK}

    # start hyper-parameters optimization

    trials = Trials()
    best_results = fmin(hyper_opt, space_, algo=tpe.suggest, max_evals=OPT_ITERS, trials=trials, show_progressbar=False)

    best_model = svm.SVC(C=best_results['C'], gamma=best_results['gamma'], kernel='rbf', random_state=1,
                         probability=True, class_weight='balanced', cache_size=2000, max_iter=10000) \
        if task_type == 'cla' else svm.SVR(C=best_results['C'], gamma=best_results['gamma'], kernel='rbf',
                                           cache_size=2000, max_iter=10000)
    best_model.fit(data_tr_x, data_tr_y)
    if model_dir :
        model_name = str(model_dir) +'/%s_%s_%s_%s'%(split_type,task_type,FP_type,'SVM_bestModel.pkl')
        joblib.dump(best_model,model_name)
    num_of_compounds = len(X)
    if task_type == 'cla':
        tr_pred = best_model.predict_proba(data_tr_x)
        tr_results = [FP_type, split_type, 'tr', num_of_compounds,
                      best_results['C'],
                      best_results['gamma']]
        tr_results.extend(statistical(data_tr_y, np.argmax(tr_pred, axis=1), tr_pred[:, 1]))
        pd_res.append(tr_results)
        # validation set
        va_pred = best_model.predict_proba(data_va_x)
        va_results = [FP_type, split_type, 'va', num_of_compounds,
                      best_results['C'],
                      best_results['gamma']]
        va_results.extend(statistical(data_va_y, np.argmax(va_pred, axis=1), va_pred[:, 1]))
        pd_res.append(va_results)
        # test set
        te_pred = best_model.predict_proba(data_te_x)
        te_results = [FP_type, split_type, 'te', num_of_compounds,
                      best_results['C'],
                      best_results['gamma']]
        te_results.extend(statistical(data_te_y, np.argmax(te_pred, axis=1), te_pred[:, 1]))
        pd_res.append(te_results)
        para_res = pd.DataFrame(pd_res, columns=['FP_type', 'split_type', 'type', 'num_of_compounds',
                                                 'para1', 'para2',
                                                 'precision', 'se', 'sp',
                                                 'acc', 'mcc', 'auc_prc', 'auc_roc'])
    else:
        # training set
        tr_pred = best_model.predict(data_tr_x)
        tr_results = [FP_type, split_type, 'tr',  num_of_compounds,
                      best_results['C'],
                      best_results['gamma'],
                      np.sqrt(mean_squared_error(data_tr_y, tr_pred)), r2_score(data_tr_y, tr_pred),
                      mean_absolute_error(data_tr_y, tr_pred)]
        pd_res.append(tr_results)
        # validation set
        va_pred = best_model.predict(data_va_x)
        va_results = [FP_type, split_type,'va',  num_of_compounds,
                      best_results['C'],
                      best_results['gamma'],
                      np.sqrt(mean_squared_error(data_va_y, va_pred)), r2_score(data_va_y, va_pred),
                      mean_absolute_error(data_va_y, va_pred)]
        pd_res.append(va_results)
        # test set
        te_pred = best_model.predict(data_te_x)
        te_results = [FP_type, split_type, 'te',  num_of_compounds,
                      best_results['C'],
                      best_results['gamma'],
                      np.sqrt(mean_squared_error(data_te_y, te_pred)), r2_score(data_te_y, te_pred),
                      mean_absolute_error(data_te_y, te_pred)]
        pd_res.append(te_results)
        para_res = pd.DataFrame(pd_res, columns=['FP_type', 'split_type', 'type', 'num_of_compounds',
                                                 'C', 'gamma',
                                                 'rmse', 'r2', 'mae'])

    pd_res = []
    for i in range(10):
        item = best_model_runing((i+1)*500,best_results,X,Y,split_type=split_type,FP_type=FP_type,task_type = task_type,model_dir=model_dir)
        pd_res.extend(item)

    if task_type =='cla':
        best_res = pd.DataFrame(pd_res, columns=['seed','FP_type', 'split_type', 'type',
                                                         'num_of_compounds', 'precision', 'se', 'sp',
                                                         'acc', 'mcc', 'auc_prc', 'auc_roc'])
    else:
        best_res = pd.DataFrame(pd_res, columns=['seed', 'FP_type', 'split_type', 'type',
                                                 'num_of_compounds', 'rmse', 'r2', 'mae'])
    return  para_res,best_res





import pandas as pd
df = pd.read_csv('/data/jianping/bokey/OCAICM/dataset/TEN/TEN_pro.csv')
cols = list(df.columns)
cols.remove('Smiles')
# df = df.loc[[1,2,3],list(df.columns).remove('Smiles')]
print(list(df.columns))


import time
import gc
from dgl.data.utils import Subset
import pandas as pd
from dgl.data.chem import csv_dataset, smiles_to_bigraph, MoleculeCSVDataset
from bokeys.gnn_utils import AttentiveFPBondFeaturizer, AttentiveFPAtomFeaturizer, collate_molgraphs, \
    EarlyStopping, set_random_seed, Meter
from torch.nn import BCEWithLogitsLoss, MSELoss
from torch.utils.data import DataLoader
import torch
from bokeys.splitdater import split_dataset
from dgl.model_zoo.chem import MPNNModel, GCNClassifier, GATClassifier, AttentiveFP
import numpy as np
from sklearn.model_selection import train_test_split
from dgl import backend as F
from hyperopt import fmin, tpe, hp, Trials
torch.set_num_threads(48)

start_time = time.time()
torch.backends.cudnn.enabled = True
torch.backends.cudnn.benchmark = True
set_random_seed(seed=43)


def get_split_index(X, Y,split_type='random'):
    data_tr_x, data_va_x, data_te_x, data_tr_y, data_va_y, data_te_y = split_dataset(X, Y, split_type=split_type,
                                                                                     valid_need=True)
    tr = pd.DataFrame([data_tr_x, data_tr_y]).T;
    tr['group'] = 'train'
    va = pd.DataFrame([data_va_x, data_va_y]).T;
    va['group'] = 'valid'
    te = pd.DataFrame([data_te_x, data_te_y]).T;
    te['group'] = 'test'
    df = pd.concat([tr, va, te])
    return df[df.group == 'train'].index, df[df.group == 'valid'].index, df[
        df.group == 'test'].index

def run_a_train_epoch(model, data_loader, loss_func, optimizer, args):
    model.train()
    train_metric = Meter()  # for each epoch
    for batch_id, batch_data in enumerate(data_loader):

        smiles, bg, labels, masks = batch_data
        atom_feats = bg.ndata.pop('h')
        bond_feats = bg.edata.pop('e')

        # transfer the data to device(cpu or cuda)
        labels, masks, atom_feats, bond_feats = labels.to(args['device']), masks.to(args['device']), atom_feats.to(
            args['device']), bond_feats.to(args['device'])

        outputs = model(bg, atom_feats) if args['model'] in ['gcn', 'gat'] else model(bg, atom_feats,
                                                                                               bond_feats)

        loss = (loss_func(outputs, labels) * (masks != 0).float()).mean()
        optimizer.zero_grad()
        loss.backward()

        optimizer.step()

        outputs.cpu()

        labels.cpu()
        masks.cpu()
        atom_feats.cpu()
        bond_feats.cpu()
        loss.cpu()
        # torch.cuda.empty_cache()

        train_metric.update(outputs, labels, masks)

    if args['metric'] == 'rmse':
        rmse_score = np.mean(train_metric.compute_metric(args['metric']))  # in case of multi-tasks
        mae_score = np.mean(train_metric.compute_metric('mae'))  # in case of multi-tasks
        r2_score = np.mean(train_metric.compute_metric('r2'))  # in case of multi-tasks
        return {'rmse': rmse_score, 'mae': mae_score, 'r2': r2_score}
    else:
        roc_score = np.mean(train_metric.compute_metric(args['metric']))  # in case of multi-tasks
        prc_score = np.mean(train_metric.compute_metric('prc_auc'))  # in case of multi-tasks

        return {'roc_auc': roc_score, 'prc_auc': prc_score}


def run_an_eval_epoch(model, data_loader, args):
    model.eval()
    eval_metric = Meter()
    with torch.no_grad():
        for batch_id, batch_data in enumerate(data_loader):
            smiles, bg, labels, masks = batch_data
            atom_feats = bg.ndata.pop('h')
            bond_feats = bg.edata.pop('e')

            # transfer the data to device(cpu or cuda)
            labels, masks, atom_feats, bond_feats = labels.to(args['device']), masks.to(args['device']), atom_feats.to(
                args['device']), bond_feats.to(args['device'])
            outputs = model(bg, atom_feats) if args['model'] in ['gcn', 'gat'] else model(bg, atom_feats,
                                                                                                   bond_feats)
            outputs.cpu()
            labels.cpu()
            masks.cpu()
            atom_feats.cpu()
            bond_feats.cpu()
            # loss.cpu()
            torch.cuda.empty_cache()
            eval_metric.update(outputs, labels, masks)
    if args['metric'] == 'rmse':
        rmse_score = np.mean(eval_metric.compute_metric(args['metric']))  # in case of multi-tasks
        mae_score = np.mean(eval_metric.compute_metric('mae'))  # in case of multi-tasks
        r2_score = np.mean(eval_metric.compute_metric('r2'))  # in case of multi-tasks
        return {'rmse': rmse_score, 'mae': mae_score, 'r2': r2_score}
    else:
        roc_score = np.mean(eval_metric.compute_metric(args['metric']))  # in case of multi-tasks
        prc_score = np.mean(eval_metric.compute_metric('prc_auc'))  # in case of multi-tasks
        se = np.mean(eval_metric.compute_metric('se'), axis=0)
        sp = np.mean(eval_metric.compute_metric('sp'), axis=0)
        acc = np.mean(eval_metric.compute_metric('acc'), axis=0)
        mcc = np.mean(eval_metric.compute_metric('mcc'),axis=0)
        precision = np.mean(eval_metric.compute_metric('precision'),axis=0)
        return {'roc_auc': roc_score, 'prc_auc': prc_score,'se':se,'sp':sp,'acc':acc,'mcc':mcc,'pre':precision}


def get_pos_weight(my_dataset):
    num_pos = F.sum(my_dataset.labels, dim=0)
    num_indices = F.tensor(len(my_dataset.labels))
    return (num_indices - num_pos) / num_pos


def all_one_zeros(series):
    if (len(series.dropna().unique()) == 2):
        flag = False
    else:
        flag = True
    return flag




def tvt_dl(X,Y,split_type='random',model_name='gcn',task_type='cla',file_name=None,model_dir=None,device ='cpu',difftasks='activity'):
    record_file = model_dir+'/dl_best_model.csv'
    f = open(record_file,'a')
    # device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

    device = device if device =='cpu' else 'cuda:0'
    # os.environ["CUDA_VISIBLE_DEVICES"] = gpu_id
    # if device == 'cuda':
    #    torch.cuda.set_device(eval(gpu_id))  # gpu device id
    file_name = file_name.replace('.csv', '_pro.csv')
    my_df = pd.read_csv(file_name)


    AtomFeaturizer = AttentiveFPAtomFeaturizer
    BondFeaturizer = AttentiveFPBondFeaturizer
    epochs = 300
    batch_size = 128
    patience = 50
    opt_iters = 50
    repetitions = 10
    num_workers = 0
    args = {'device': device, 'task': difftasks,
            'metric': 'roc_auc' if task_type == 'cla' else 'rmse', 'model': model_name}
    tasks = args['task']

    Hspace = {'gcn': dict(l2=hp.choice('l2', [0, 10 ** -8, 10 ** -6, 10 ** -4]),
                          lr=hp.choice('lr', [10 ** -2.5, 10 ** -3.5, 10 ** -1.5]),
                          gcn_hidden_feats=hp.choice('gcn_hidden_feats',
                                                     [[128, 128], [256, 256], [128, 64], [256, 128]]),
                          classifier_hidden_feats=hp.choice('classifier_hidden_feats', [128, 64, 256])),
              'mpnn': dict(l2=hp.choice('l2', [0, 10 ** -8, 10 ** -6, 10 ** -4]),
                           lr=hp.choice('lr', [10 ** -2.5, 10 ** -3.5, 10 ** -1.5]),
                           node_hidden_dim=hp.choice('node_hidden_dim', [64, 32, 16]),
                           edge_hidden_dim=hp.choice('edge_hidden_dim', [64, 32, 16]),
                           num_layer_set2set=hp.choice('num_layer_set2set', [2, 3, 4])),
              'gat': dict(l2=hp.choice('l2', [0, 10 ** -8, 10 ** -6, 10 ** -4]),
                          lr=hp.choice('lr', [10 ** -2.5, 10 ** -3.5, 10 ** -1.5]),
                          gat_hidden_feats=hp.choice('gat_hidden_feats',
                                                     [[128, 128], [256, 256], [128, 64], [256, 128]]),
                          num_heads=hp.choice('num_heads', [[2, 2], [3, 3], [4, 4], [4, 3], [3, 2]]),
                          classifier_hidden_feats=hp.choice('classifier_hidden_feats', [128, 64, 256])),
              'attentivefp': dict(l2=hp.choice('l2', [0, 10 ** -8, 10 ** -6, 10 ** -4]),
                                  lr=hp.choice('lr', [10 ** -2.5, 10 ** -3.5, 10 ** -1.5]),
                                  num_layers=hp.choice('num_layers', [2, 3, 4, 5, 6]),
                                  num_timesteps=hp.choice('num_timesteps', [1, 2, 3, 4, 5]),
                                  dropout=hp.choice('dropout', [ 0.1, 0.3, 0.5]),
                                  graph_feat_size=hp.choice('graph_feat_size', [50, 100, 200, 300]))}
    hyper_space = Hspace[args['model']]


    my_dataset: MoleculeCSVDataset = csv_dataset.MoleculeCSVDataset(my_df.iloc[:, :], smiles_to_bigraph, AtomFeaturizer,
                                                                    BondFeaturizer, 'Smiles',
                                                                    file_name.replace('.csv', '.bin'))
    if task_type == 'cla':
        pos_weight = get_pos_weight(my_dataset)
    else:
        pos_weight = None





    tr_indx, val_indx, te_indx = get_split_index(X,Y,split_type=split_type)
    train_loader = DataLoader(Subset(my_dataset, tr_indx), batch_size=batch_size, shuffle=True,
                              collate_fn=collate_molgraphs, num_workers=num_workers)
    val_loader = DataLoader(Subset(my_dataset, val_indx), batch_size=batch_size, shuffle=True,
                            collate_fn=collate_molgraphs, num_workers=num_workers)
    test_loader = DataLoader(Subset(my_dataset, te_indx), batch_size=batch_size, shuffle=True,
                             collate_fn=collate_molgraphs, num_workers=num_workers)


    def hyper_opt(hyper_paras):
        # get the model instance
        if model_name == 'gcn':
            my_model = GCNClassifier(in_feats=AtomFeaturizer.feat_size('h'),
                                     gcn_hidden_feats=hyper_paras['gcn_hidden_feats'],
                                     n_tasks=len(tasks), classifier_hidden_feats=hyper_paras['classifier_hidden_feats'])
            model_file_name = '%s/%s_%s_%s_%s_%.6f_%s_%s.pth' % (model_dir, args['model'], split_type, args['task'],
                                                                 hyper_paras['l2'], hyper_paras['lr'],
                                                                 hyper_paras['gcn_hidden_feats'],
                                                                 hyper_paras['classifier_hidden_feats'])

        elif model_name == 'mpnn':
            my_model = MPNNModel(node_input_dim=AtomFeaturizer.feat_size('h'), edge_input_dim=BondFeaturizer.feat_size('e'),
                                 output_dim=len(tasks), node_hidden_dim=hyper_paras['node_hidden_dim'],
                                 edge_hidden_dim=hyper_paras['edge_hidden_dim'],
                                 num_layer_set2set=hyper_paras['num_layer_set2set'])
            model_file_name = '%s/%s_%s_%s_%s_%.6f_%s_%s_%s.pth' % (model_dir, args['model'], split_type, args['task'],
                                                                    hyper_paras['l2'], hyper_paras['lr'],
                                                                    hyper_paras['node_hidden_dim'],
                                                                    hyper_paras['edge_hidden_dim'],
                                                                    hyper_paras['num_layer_set2set'])
        elif model_name == 'attentivefp':
            my_model = AttentiveFP(node_feat_size=AtomFeaturizer.feat_size('h'),
                                   edge_feat_size=BondFeaturizer.feat_size('e'),
                                   num_layers=hyper_paras['num_layers'], num_timesteps=hyper_paras['num_timesteps'],
                                   graph_feat_size=hyper_paras['graph_feat_size'], output_size=len(tasks),
                                   dropout=hyper_paras['dropout'])
            model_file_name = '%s/%s_%s_%s_%s_%.6f_%s_%s_%s_%s.pth' % (
            model_dir, args['model'], split_type, args['task'],
            hyper_paras['l2'], hyper_paras['lr'],
            hyper_paras['num_layers'],
            hyper_paras['num_timesteps'],
            hyper_paras['graph_feat_size'],
            hyper_paras['dropout'])

        else:
            my_model = GATClassifier(in_feats=AtomFeaturizer.feat_size('h'),
                                     gat_hidden_feats=hyper_paras['gat_hidden_feats'],
                                     num_heads=hyper_paras['num_heads'], n_tasks=len(tasks),
                                     classifier_hidden_feats=hyper_paras['classifier_hidden_feats'])
            model_file_name = '%s/%s_%s_%s_%s_%.6f_%s_%s_%s.pth' % (model_dir, args['model'], split_type, args['task'],
                                                                    hyper_paras['l2'], hyper_paras['lr'],
                                                                    hyper_paras['gat_hidden_feats'],
                                                                    hyper_paras['num_heads'],
                                                                    hyper_paras['classifier_hidden_feats'])

        optimizer = torch.optim.Adam(my_model.parameters(), lr=hyper_paras['lr'], weight_decay=hyper_paras['l2'])

        if task_type == 'reg':
            loss_func = MSELoss(reduction='none')
            stopper = EarlyStopping(mode='lower', patience=patience,filename=model_file_name)
        else:
            loss_func = BCEWithLogitsLoss(reduction='none', pos_weight=pos_weight.to(args['device']))
            stopper = EarlyStopping(mode='higher', patience=patience,filename=model_file_name)
        my_model.to(device)

        for j in range(epochs):
            # training
            run_a_train_epoch(my_model, train_loader, loss_func, optimizer, args)

            # early stopping
            val_scores = run_an_eval_epoch(my_model, val_loader, args)
            early_stop = stopper.step(val_scores[args['metric']], my_model)

            if early_stop:
                break
        stopper.load_checkpoint(my_model)
        # tr_scores = run_an_eval_epoch(my_model, train_loader, args)
        val_scores = run_an_eval_epoch(my_model, val_loader, args)
        # te_scores = run_an_eval_epoch(my_model, test_loader, args)

        feedback = val_scores[args['metric']] if task_type == 'reg' else (1 - val_scores[args['metric']])
        my_model.cpu()
        torch.cuda.empty_cache()
        gc.collect()
        return feedback


    # start hyper-parameters optimization

    trials = Trials()
    opt_res = fmin(hyper_opt, hyper_space, algo=tpe.suggest, max_evals=opt_iters, trials=trials)

   # construct the model based on the optimal hyper-parameters
    l2_ls = [0, 10 ** -8, 10 ** -6, 10 ** -4]
    lr_ls = [10 ** -2.5, 10 ** -3.5, 10 ** -1.5]
    hidden_feats_ls = [(128, 128), (256, 256), (128, 64), (256, 128)]
    classifier_hidden_feats_ls = [128, 64, 256]
    node_hidden_dim_ls = [64, 32, 16]
    edge_hidden_dim_ls = [64, 32, 16]
    num_layer_set2set_ls = [2, 3, 4]
    num_heads_ls = [(2, 2), (3, 3), (4, 4), (4, 3), (3, 2)]
    num_layers_ls = [2, 3, 4, 5, 6]
    num_timesteps_ls = [1, 2, 3, 4, 5]
    graph_feat_size_ls = [50, 100, 200, 300]
    dropout_ls = [ 0.1, 0.3, 0.5]
    if model_name == 'gcn':
        best_model = GCNClassifier(in_feats=AtomFeaturizer.feat_size('h'),
                                   gcn_hidden_feats=hidden_feats_ls[opt_res['gcn_hidden_feats']],
                                   n_tasks=len(tasks),
                                   classifier_hidden_feats=classifier_hidden_feats_ls[opt_res['classifier_hidden_feats']])
        best_model_file = '%s/%s_%s_%s_%s_%.6f_%s_%s.pth' % (model_dir,args['model'], split_type,args['task'],
                                                                         l2_ls[opt_res['l2']], lr_ls[opt_res['lr']],
                                                                         hidden_feats_ls[opt_res['gcn_hidden_feats']],
                                                                         classifier_hidden_feats_ls[opt_res['classifier_hidden_feats']])
        f.write('{}\n'.format(best_model_file))
        param = {'l2':l2_ls[opt_res['l2']], 'lr':lr_ls[opt_res['lr']],
             'gcn_hidden_feat':hidden_feats_ls[opt_res['gcn_hidden_feats']],
             'classifier_hidden_feats':classifier_hidden_feats_ls[opt_res['classifier_hidden_feats']]}
    elif model_name == 'mpnn':
        best_model = MPNNModel(node_input_dim=AtomFeaturizer.feat_size('h'), edge_input_dim=BondFeaturizer.feat_size('e'),
                               output_dim=len(tasks), node_hidden_dim=node_hidden_dim_ls[opt_res['node_hidden_dim']],
                               edge_hidden_dim=edge_hidden_dim_ls[opt_res['edge_hidden_dim']],
                               num_layer_set2set=num_layer_set2set_ls[opt_res['num_layer_set2set']])
        best_model_file = '%s/%s_%s_%s_%s_%.6f_%s_%s_%s.pth' % (model_dir,args['model'], split_type, args['task'],
                                                                        l2_ls[opt_res['l2']], lr_ls[opt_res['lr']],
                                                                        node_hidden_dim_ls[opt_res['node_hidden_dim']],
                                                                        edge_hidden_dim_ls[opt_res['edge_hidden_dim']],
                                                                        num_layer_set2set_ls[opt_res['num_layer_set2set']])
        param = {'l2': l2_ls[opt_res['l2']], 'lr': lr_ls[opt_res['lr']],
                 'node_hidden_dim': node_hidden_dim_ls[opt_res['node_hidden_dim']],
                 'edge_hidden_dim': edge_hidden_dim_ls[opt_res['edge_hidden_dim']],
                 'num_layer_set2set': num_layer_set2set_ls[opt_res['num_layer_set2set']]}
        f.write('{}\n'.format(best_model_file))
    elif model_name == 'attentivefp':
        best_model = AttentiveFP(node_feat_size=AtomFeaturizer.feat_size('h'), edge_feat_size=BondFeaturizer.feat_size('e'),
                               num_layers=num_layers_ls[opt_res['num_layers']],
                               num_timesteps=num_timesteps_ls[opt_res['num_timesteps']],
                               graph_feat_size=graph_feat_size_ls[opt_res['graph_feat_size']], output_size=len(tasks),
                               dropout=dropout_ls[opt_res['dropout']])
        best_model_file = '%s/%s_%s_%s_%s_%.6f_%s_%s_%s_%s.pth' % (model_dir,args['model'], split_type, args['task'],
                                                                           l2_ls[opt_res['l2']], lr_ls[opt_res['lr']],
                                                                                   num_layers_ls[opt_res['num_layers']],
                                                                                   num_timesteps_ls[opt_res['num_timesteps']],
                                                                                   graph_feat_size_ls[opt_res['graph_feat_size']],
                                                                                   dropout_ls[opt_res['dropout']])
        f.write('{}\n'.format(best_model_file))
        param = {'l2': l2_ls[opt_res['l2']], 'lr': lr_ls[opt_res['lr']],
                 'num_layers': num_layers_ls[opt_res['num_layers']],
                 'num_timesteps': num_timesteps_ls[opt_res['num_timesteps']],
                 'graph_feat_size': graph_feat_size_ls[opt_res['graph_feat_size']],
                 'dropout':dropout_ls[opt_res['dropout']]}

    else:
        best_model = GATClassifier(in_feats=AtomFeaturizer.feat_size('h'),
                                   gat_hidden_feats=hidden_feats_ls[opt_res['gat_hidden_feats']],
                                   num_heads=num_heads_ls[opt_res['num_heads']], n_tasks=len(tasks),
                                   classifier_hidden_feats=classifier_hidden_feats_ls[opt_res['classifier_hidden_feats']])
        best_model_file = '%s/%s_%s_%s_%s_%.6f_%s_%s_%s.pth' % (model_dir,args['model'], split_type, args['task'],
                                                                        l2_ls[opt_res['l2']], lr_ls[opt_res['lr']],
                                                                                hidden_feats_ls[opt_res['gat_hidden_feats']],
                                                                                num_heads_ls[opt_res['num_heads']],
                                                                                classifier_hidden_feats_ls[
                                                                                opt_res['classifier_hidden_feats']])
        f.write('{}\n'.format(best_model_file))
        param = {'l2': l2_ls[opt_res['l2']], 'lr': lr_ls[opt_res['lr']],
                 'gat_hidden_feats': hidden_feats_ls[opt_res['gat_hidden_feats']],
                 'num_heads': num_heads_ls[opt_res['num_heads']],
                 'classifier_hidden_feats': classifier_hidden_feats_ls[opt_res['classifier_hidden_feats']]}
    f.close()
    best_model.load_state_dict(torch.load(best_model_file, map_location=device)['model_state_dict'])
    best_model.to(device)
    tr_scores = run_an_eval_epoch(best_model, train_loader, args)
    val_scores = run_an_eval_epoch(best_model, val_loader, args)
    te_scores = run_an_eval_epoch(best_model, test_loader, args)

    record = {'roc_auc':[tr_scores['roc_auc'],val_scores['roc_auc'],te_scores['roc_auc']],
              'prc_auc':[tr_scores['prc_auc'],val_scores['prc_auc'],te_scores['prc_auc']],
              'se':[tr_scores['se'],val_scores['se'],te_scores['se']],
              'sp':[tr_scores['sp'],val_scores['sp'],te_scores['sp']],
              'acc':[tr_scores['acc'],val_scores['acc'],te_scores['acc']],
              'mcc':[tr_scores['mcc'],val_scores['mcc'],te_scores['mcc']],
              'precision':[tr_scores['pre'],val_scores['pre'],te_scores['pre']]}
    param = {k:[v,v,v] for k,v in param.items()}
    record = {k:v for d in [param,record] for k,v in d.items()}
    best_res = pd.DataFrame(record,index=['tr','val','te'])

    best_res['split'] = split_type
    best_res['model'] = model_name





    tr_res = []
    val_res = []
    te_res = []


    for seed in range(1, repetitions + 1):

        seed = seed*500


        if args['metric'] == 'roc_auc':

            while True:
                data_tr,data_re  = train_test_split(my_df, test_size=0.2, random_state=seed)

                data_va, data_te = train_test_split(data_re, test_size=0.5, random_state=seed)
                if np.any(data_tr[tasks].apply(all_one_zeros)) or \
                        np.any(data_va[tasks].apply(all_one_zeros)) or \
                        np.any(data_te[tasks].apply(all_one_zeros)):

                    seed = np.random.randint(50, 999999)
                else:
                    pass
                    break
        else:
            data_tr, data_re = train_test_split(my_df, test_size=0.2, random_state=seed)

            data_va, data_te = train_test_split(data_re, test_size=0.5, random_state=seed)
        tr_indx, val_indx, te_indx = data_tr.index, data_va.index, data_te.index
        train_loader = DataLoader(Subset(my_dataset, tr_indx), batch_size=batch_size, shuffle=True,
                                  collate_fn=collate_molgraphs, num_workers=num_workers)
        val_loader = DataLoader(Subset(my_dataset, val_indx), batch_size=batch_size, shuffle=True,
                                collate_fn=collate_molgraphs, num_workers=num_workers)
        test_loader = DataLoader(Subset(my_dataset, te_indx), batch_size=batch_size, shuffle=True,
                                 collate_fn=collate_molgraphs, num_workers=num_workers)
        best_model_file = '%s/%s_%s_%s_bst_%s.pth' % (model_dir,args['model'], split_type, args['task'], seed)

        if model_name == 'gcn':
            best_model = GCNClassifier(in_feats=AtomFeaturizer.feat_size('h'),
                                       gcn_hidden_feats=hidden_feats_ls[opt_res['gcn_hidden_feats']],
                                       n_tasks=len(tasks),
                                       classifier_hidden_feats=classifier_hidden_feats_ls[
                                           opt_res['classifier_hidden_feats']])

        elif model_name == 'gat':
            best_model = GATClassifier(in_feats=AtomFeaturizer.feat_size('h'),
                                       gat_hidden_feats=hidden_feats_ls[opt_res['gat_hidden_feats']],
                                       num_heads=num_heads_ls[opt_res['num_heads']], n_tasks=len(tasks),
                                       classifier_hidden_feats=classifier_hidden_feats_ls[
                                           opt_res['classifier_hidden_feats']])
        elif model_name == 'attentivefp':
            best_model = AttentiveFP(node_feat_size=AtomFeaturizer.feat_size('h'),
                                     edge_feat_size=BondFeaturizer.feat_size('e'),
                                     num_layers=num_layers_ls[opt_res['num_layers']],
                                     num_timesteps=num_timesteps_ls[opt_res['num_timesteps']],
                                     graph_feat_size=graph_feat_size_ls[opt_res['graph_feat_size']], output_size=len(tasks),
                                     dropout=dropout_ls[opt_res['dropout']])
        else:
            best_model = MPNNModel(node_input_dim=AtomFeaturizer.feat_size('h'),
                                   edge_input_dim=BondFeaturizer.feat_size('e'),
                                   output_dim=len(tasks), node_hidden_dim=node_hidden_dim_ls[opt_res['node_hidden_dim']],
                                   edge_hidden_dim=edge_hidden_dim_ls[opt_res['edge_hidden_dim']],
                                   num_layer_set2set=num_layer_set2set_ls[opt_res['num_layer_set2set']])

        best_optimizer = torch.optim.Adam(best_model.parameters(), lr=lr_ls[opt_res['lr']],
                                          weight_decay=l2_ls[opt_res['l2']])
        if task_type == 'reg':
            loss_func = MSELoss(reduction='none')
            stopper = EarlyStopping(mode='lower', patience=patience, filename=best_model_file)
        else:
            loss_func = BCEWithLogitsLoss(reduction='none', pos_weight=pos_weight.to(args['device']))
            stopper = EarlyStopping(mode='higher', patience=patience,filename=best_model_file)
        best_model.to(device)

        for j in range(epochs):
            # training
            st = time.time()
            run_a_train_epoch(best_model, train_loader, loss_func, best_optimizer, args)
            end = time.time()
            # early stopping
            train_scores = run_an_eval_epoch(best_model, train_loader, args)
            val_scores = run_an_eval_epoch(best_model, val_loader, args)
            early_stop = stopper.step(val_scores[args['metric']], best_model)
            if early_stop:
                break
        stopper.load_checkpoint(best_model)
        tr_scores = run_an_eval_epoch(best_model, train_loader, args)
        val_scores = run_an_eval_epoch(best_model, val_loader, args)
        te_scores = run_an_eval_epoch(best_model, test_loader, args)

        tr_res.append(tr_scores);
        val_res.append(val_scores);
        te_res.append(te_scores)
    if task_type == 'reg':
        cols = ['rmse', 'mae', 'r2']
    else:
        cols = ['auc_roc', 'auc_prc','se','sp','acc','mcc','precision']
    tr = [list(item.values()) for item in tr_res]
    val = [list(item.values()) for item in val_res]
    te = [list(item.values()) for item in te_res]
    tr_pd = pd.DataFrame(tr, columns=cols)
    tr_pd['seed'] = [seed*500 for seed in range(1, repetitions + 1)]
    tr_pd['type'] = 'tr'
    val_pd = pd.DataFrame(val, columns=cols)
    val_pd['seed'] = [seed*500 for seed in range(1, repetitions + 1)]
    val_pd['type'] = 'va'
    te_pd = pd.DataFrame(te, columns=cols)
    te_pd['seed'] = [seed*500 for seed in range(1, repetitions + 1)]
    te_pd['type'] = 'te'
    sta_pd = pd.concat([tr_pd, val_pd, te_pd], ignore_index=True)
    sta_pd['model'] = args['model']

    sta_pd['split'] = split_type



    return best_res,sta_pd



import pandas as pd
import numpy as np
from rdkit.Chem import AllChem
def create_des(X,Y ,FP_type = 'ECFP4',model_dir=False):
    smiles = np.array(X)
    if FP_type == 'ECFP4':
        ms = [AllChem.MolFromSmiles(smiles[i]) for i in range(len(smiles))]
        ecfpMat = np.zeros((len(ms), 1024), dtype=int)
        for i in range(len(ms)):
            fp = AllChem.GetMorganFingerprintAsBitVect(ms[i], 2, 1024)
            ecfpMat[i] = np.array(list(fp.ToBitString()))
        X = ecfpMat
    elif FP_type == 'MACCS':
        ms = [AllChem.MolFromSmiles(smiles[i]) for i in range(len(smiles))]
        ecfpMat = np.zeros((len(ms), 167), dtype=int)
        for i in range(len(ms)):
            fp = AllChem.GetMACCSKeysFingerprint(ms[i])
            ecfpMat[i] = np.array(list(fp.ToBitString()))
        X = ecfpMat
    elif FP_type == '2d-3d':
        dataset = str(model_dir).split('/model_save')[0].split('/')[-1]
        des_file = str(model_dir).split('/model_save')[0] +'/'+dataset+'_23d_adj.csv'
        des = pd.read_csv(des_file).drop_duplicates()
        des = des.set_index('Name')
        des = des.dropna(axis=1)
        features = len(des.columns)
        ecfpMat = np.zeros((len(X), features), dtype=float)
        for j,smile in enumerate(smiles):
            index = str(smile)
            try:
                ecfpMat[j] = np.array(des.loc[index,:])
            except:
                pass
        X = ecfpMat
    elif FP_type == 'pubchem':
        dataset = str(model_dir).split('/model_save')[0].split('/')[-1]
        des_file = str(model_dir).split('/model_save')[0] + '/' + dataset + '_pubchem.csv'
        des = pd.read_csv(des_file).drop_duplicates()
        des = des.set_index('Name')
        features = len(des.columns)
        ecfpMat = np.zeros((len(X), features), dtype=int)
        for j, smile in enumerate(smiles):
            index = str(smile)
            try:
                ecfpMat[j] = np.array(des.loc[index, :])
            except:
                pass
        X = ecfpMat
    return X,Y


import setuptools

setuptools.setup(
        name='bokeys',  # 包名
        version='0.1.5',  # 版本
        description="",  # 包简介
        long_description=open('README.md').read(),  # 读取文件中介绍包的详细内容
        include_package_data=True,  # 是否允许上传资源文件
        author='bokey',  # 作者
        author_email='',  # 作者邮件
        url='',  # github或者自己的网站地址
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'Topic :: Software Development :: Build Tools',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3',  # 设置编写时的python版本
        ],
        python_requires='>=3.8',  # 设置python版本要求
        install_requires=[''],  # 安装所需要的库
        entry_points={
            'console_scripts': [
                ''],
        },  # 设置命令行工具(可不使用就可以注释掉)
)



import os
import shutil
from bokeys import compute_info
from bokeys import data_pre
from bokeys import run_model
import time



def running(file_name,out_dir = './',split=None,model=None,FP=None, cpus=4):
    split = split if type(split) != type(None) else ['random']
    model = model if type(model) != type(None) else ['SVM']
    FP = FP if type(FP) != type(None) else ['MACCS']
    FP_list = ['2d-3d','MACCS','ECFP4','pubchem']
    split_list = ['random', 'scaffold', 'cluster']
    model_list = ['SVM','KNN','DNN','RF','XGB','gcn','gat','attentivefp','mpnn']
    assert len(set(split) - set(split_list)) == 0, '{} element need in {}'.format(split, split_list)
    assert len(set(model)-set(model_list))==0,'{} element need in {}'.format(model ,model_list)
    assert len(set(FP) - set(FP_list)) == 0, '{} element need in {}'.format(FP, FP_list)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    assert len(set(list([file_name.split('.')[-1]])) -set(list(['csv','txt','tsv'])))==0, '{} need in ["csv","txt","tsv"]'.format(file_name.split('.')[-1])
    dataset = file_name.split('/')[-1].split('.'+file_name.split('.')[-1])[0]
    dir_path = os.path.join(out_dir, dataset)
    if not os.path.exists(dir_path):
        os.mkdir(dir_path)
    file = os.path.join(dir_path, dataset + '.csv')
    shutil.copy(file_name, os.path.join(dir_path,dataset + '.csv'))
    des = [fp for fp in FP if fp in ['pubchem','2d-3d']]
    data_pre.start(file, des=des)
    mpl = True if cpus>1 else False
    run_model.main(file, split, FP, model, cpus, mpl, 'cpu')
    compute_info.get_info(file_name=file, models=model, split=split, FP=FP)
    compute_info.para(file_name=file)

if __name__ == '__main__':
    running('/data/jianping/bokey/OCAICM/dataset/GLS1/GLS1.csv')




import torch
import numpy as np
import pandas as pd
from bokeys.dnn_torch_utils import Meter, MyDataset, EarlyStopping, MyDNN, collate_fn, set_random_seed
from hyperopt import fmin, tpe, hp, rand, STATUS_OK, Trials, partial
from torch.utils.data import DataLoader
from torch.nn import BCEWithLogitsLoss, MSELoss
import gc
import time
start_time = time.time()
import warnings
from sklearn import preprocessing
from bokeys.splitdater import split_dataset
from bokeys.feature_create  import create_des
warnings.filterwarnings('ignore')
torch.backends.cudnn.enabled = True
torch.backends.cudnn.benchmark = True
set_random_seed(seed=43)
torch.set_num_threads(48)
def get_split_index(X, Y,split_type='random',random =20):
    data_tr_x, data_va_x, data_te_x, data_tr_y, data_va_y, data_te_y = split_dataset(X, Y, split_type=split_type,
                                                                                     valid_need=True,random_state=random)
    tr = pd.DataFrame([data_tr_x, data_tr_y]).T;
    tr['group'] = 'train'
    va = pd.DataFrame([data_va_x, data_va_y]).T;
    va['group'] = 'valid'
    te = pd.DataFrame([data_te_x, data_te_y]).T;
    te['group'] = 'test'
    df = pd.concat([tr, va, te])
    return df[df.group == 'train'].index, df[df.group == 'valid'].index, df[
        df.group == 'test'].index

def run_a_train_epoch(model, data_loader, loss_func, optimizer, args):
    model.train()

    train_metric = Meter()  # for each epoch
    for batch_id, batch_data in enumerate(data_loader):
        Xs, Ys, masks = batch_data

        # transfer the data to device(cpu or cuda)
        Xs, Ys, masks = Xs.to(args['device']), Ys.to(args['device']), masks.to(args['device'])

        outputs = model(Xs)

        loss = (loss_func(outputs, Ys) * (masks != 0).float()).mean()

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        outputs.cpu()
        Ys.cpu()
        masks.cpu()
        loss.cpu()
#        torch.cuda.empty_cache()

        train_metric.update(outputs, Ys, masks)
    if args['reg']:
        rmse_score = np.mean(train_metric.compute_metric(args['metric']))  # in case of multi-tasks
        mae_score = np.mean(train_metric.compute_metric('mae'))  # in case of multi-tasks
        r2_score = np.mean(train_metric.compute_metric('r2'))  # in case of multi-tasks
        return {'rmse': rmse_score, 'mae': mae_score, 'r2': r2_score}
    else:
        roc_score = np.mean(train_metric.compute_metric(args['metric']))  # in case of multi-tasks
        prc_score = np.mean(train_metric.compute_metric('prc_auc'))  # in case of multi-tasks
        return {'roc_auc': roc_score, 'prc_auc': prc_score}


def run_an_eval_epoch(model, data_loader, args):
    model.eval()

    eval_metric = Meter()
    with torch.no_grad():
        for batch_id, batch_data in enumerate(data_loader):
            Xs, Ys, masks = batch_data
            # transfer the data to device(cpu or cuda)
            Xs, Ys, masks = Xs.to(args['device']), Ys.to(args['device']), masks.to(args['device'])

            outputs = model(Xs)

            outputs.cpu()
            Ys.cpu()
            masks.cpu()
#            torch.cuda.empty_cache()
            eval_metric.update(outputs, Ys, masks)
    if args['reg']:
        rmse_score = np.mean(eval_metric.compute_metric(args['metric']))  # in case of multi-tasks
        mae_score = np.mean(eval_metric.compute_metric('mae'))  # in case of multi-tasks
        r2_score = np.mean(eval_metric.compute_metric('r2'))  # in case of multi-tasks
        return {'rmse': rmse_score, 'mae': mae_score, 'r2': r2_score}
    else:
        roc_score = np.mean(eval_metric.compute_metric(args['metric']))  # in case of multi-tasks
        prc_score = np.mean(eval_metric.compute_metric('prc_auc'))  # in case of multi-tasks
        se = np.mean(eval_metric.compute_metric('se'), axis=0)
        sp = np.mean(eval_metric.compute_metric('sp'), axis=0)
        acc = np.mean(eval_metric.compute_metric('acc'), axis=0)
        mcc = np.mean(eval_metric.compute_metric('mcc'), axis=0)
        precision = np.mean(eval_metric.compute_metric('precision'), axis=0)
        return {'roc_auc': roc_score, 'prc_auc': prc_score, 'se': se, 'sp': sp, 'acc': acc, 'mcc': mcc,
                'pre': precision}


def get_pos_weight(Ys):
    Ys = torch.tensor(np.nan_to_num(Ys), dtype=torch.float32)
    num_pos = torch.sum(Ys, dim=0)
    num_indices = torch.tensor(len(Ys))
    return (num_indices - num_pos) / num_pos

def standardize(col):
    return (col - np.mean(col)) / np.std(col)


def all_one_zeros(series):
    if (len(series.dropna().unique()) == 2):
        flag = False
    else:
        flag = True
    return flag



def best_model_runing(seed, opt_res, X, Y, split_type='random', FP_type='ECFP4',task_type='cla',model_dir=False,difftasks=['activity'],my_df = pd.DataFrame()):
    # construct the model based on the optimal hyper-parameters
    hidden_unit1_ls = [64, 128, 256, 512]
    hidden_unit2_ls = [64, 128, 256, 512]
    hidden_unit3_ls = [64, 128, 256, 512]
    opt_hidden_units = [hidden_unit1_ls[opt_res['hidden_unit1']], hidden_unit2_ls[opt_res['hidden_unit2']],
                        hidden_unit3_ls[opt_res['hidden_unit3']]]


    pd_res = []
    epochs = 300  # training epoch

    batch_size = 128
    patience = 50
    #device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
    device = torch.device("cpu")
    # os.environ["CUDA_VISIBLE_DEVICES"] = gpu_id
    # if device == 'cuda':
    #    torch.cuda.set_device(eval(gpu_id))  # gpu device id
    reg = False if task_type=='cla' else True

    args = {'device': device, 'metric': 'rmse' if reg else 'roc_auc', 'epochs': epochs,
            'patience': patience,  'reg': reg}
    if task_type=='cla':
        while True:
            tr_indx, val_indx, te_indx = get_split_index(X, Y, split_type='random', random=seed)

            data_tr_x, data_tr_y = my_df.loc[tr_indx, 'Smiles'], my_df.loc[tr_indx, difftasks]
            data_va_x, data_va_y = my_df.loc[tr_indx, 'Smiles'], my_df.loc[val_indx, difftasks]
            data_te_x, data_te_y = my_df.loc[tr_indx, 'Smiles'], my_df.loc[te_indx, difftasks]


            data_tr_x, data_tr_y = create_des(data_tr_x, data_tr_y, FP_type=FP_type,model_dir=model_dir)
            data_va_x, data_va_y = create_des(data_va_x, data_va_y, FP_type=FP_type,model_dir=model_dir)
            data_te_x, data_te_y = create_des(data_te_x, data_te_y, FP_type=FP_type,model_dir=model_dir)
            if np.any(data_tr_y[difftasks].apply(all_one_zeros)) or \
                        np.any(data_va_y[difftasks].apply(all_one_zeros)) or \
                        np.any(data_te_y[difftasks].apply(all_one_zeros)):
                # print('\ninvalid random seed {} due to one class presented in the splitted {} sets...'.format(seed,
                #                                                                                               split_type))
                # print('Changing to another random seed...\n')
                seed = np.random.randint(50, 999999)
            else:

                break
    else:
        data_tr_x, data_va_x, data_te_x, data_tr_y, data_va_y, data_te_y = split_dataset(X, Y,
                                                                                         split_type=split_type,
                                                                                         valid_need=True,
                                                                                         random_state=seed)
        data_tr_x, data_tr_y = create_des(data_tr_x, data_tr_y, FP_type=FP_type)
        data_va_x, data_va_y = create_des(data_va_x, data_va_y, FP_type=FP_type)
        data_te_x, data_te_y = create_des(data_te_x, data_te_y, FP_type=FP_type)

    data_tr_y = data_tr_y.values.reshape(-1, len(difftasks))
    data_va_y = data_va_y.values.reshape(-1, len(difftasks))
    data_te_y = data_te_y.values.reshape(-1, len(difftasks))



    # dataloader
    train_dataset = MyDataset(data_tr_x, data_tr_y)
    validation_dataset = MyDataset(data_va_x, data_va_y)
    test_dataset = MyDataset(data_te_x, data_te_y)
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    validation_loader = DataLoader(validation_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    inputs = data_tr_x.shape[1]
    best_model = MyDNN(inputs=inputs, hideen_units=opt_hidden_units, outputs=len(difftasks),
                       dp_ratio=opt_res['dropout'], reg=reg)

    best_optimizer = torch.optim.Adadelta(best_model.parameters(), weight_decay=opt_res['l2'])
    file_nam = '%s/%s_%s_%s_%s_%.4f_%d_%d_%d_%.4f_early_stop_%d.pth' % (
    model_dir,task_type,FP_type, split_type, 'dataset', opt_res['dropout'],
    hidden_unit1_ls[opt_res['hidden_unit1']],
    hidden_unit1_ls[opt_res['hidden_unit2']],
    hidden_unit1_ls[opt_res['hidden_unit3']],
    opt_res['l2'], seed)
    if model_dir:
        record_file = model_dir + '/dl_best_model.csv'
        f = open(record_file, 'a')
        f.write('{}\n'.format(file_nam))
        f.close()
    if reg:
        loss_func = MSELoss(reduction='none')
        stopper = EarlyStopping(mode='lower', patience=patience, filename=file_nam)
    else:
        pos_weights = get_pos_weight(Y.values)
        loss_func = BCEWithLogitsLoss(reduction='none', pos_weight=pos_weights.to(args['device']))
        stopper = EarlyStopping(mode='higher', patience=patience, filename=file_nam)
    best_model.to(device)

    for j in range(epochs):
        # training

        run_a_train_epoch(best_model, train_loader, loss_func, best_optimizer, args)

        # early stopping
        train_scores = run_an_eval_epoch(best_model, train_loader, args)
        val_scores = run_an_eval_epoch(best_model, validation_loader, args)
        early_stop = stopper.step(val_scores[args['metric']], best_model)
        if early_stop:
            break

        stopper.load_checkpoint(best_model)
    if task_type=='cla':
        tr_scores = run_an_eval_epoch(best_model, train_loader, args)
        tr_results = [seed,FP_type, split_type, 'tr',tr_scores['se'], tr_scores['sp'],
                      tr_scores['acc'],tr_scores['mcc'], tr_scores['pre'],
                      tr_scores['prc_auc'],tr_scores['roc_auc']]

        val_scores = run_an_eval_epoch(best_model, validation_loader, args)
        va_results = [seed,FP_type, split_type, 'val',val_scores['se'], val_scores['sp'],
                      val_scores['acc'],val_scores['mcc'], val_scores['pre'],
                      val_scores['prc_auc'],val_scores['roc_auc']]
        te_scores = run_an_eval_epoch(best_model, test_loader, args)
        te_results = [seed,FP_type, split_type, 'te',te_scores['se'], te_scores['sp'],
                      te_scores['acc'],te_scores['mcc'], te_scores['pre'],
                      te_scores['prc_auc'],te_scores['roc_auc']]
        pd_res.append(tr_results);pd_res.append(va_results), pd_res.append(te_results)
    else:
        tr_scores = run_an_eval_epoch(best_model, train_loader, args)
        val_scores = run_an_eval_epoch(best_model, validation_loader, args)
        te_scores = run_an_eval_epoch(best_model, test_loader, args)
        tr_results = [seed,FP_type, split_type, 'tr',
                      tr_scores['rmse'], tr_scores['mae'],
                      tr_scores['r2']]
        va_results = [seed,FP_type, split_type, 'va',
                      val_scores['rmse'], val_scores['mae'],
                      val_scores['r2']]
        te_results = [seed,FP_type, split_type, 'te',
                     te_scores['rmse'], te_scores['mae'],
                      te_scores['r2']]

        pd_res.append(tr_results); pd_res.append(va_results); pd_res.append(te_results)


    return pd_res

def tvt_dnn(X,Y,split_type='random',FP_type='ECFP4',task_type='cla',model_dir=False,file_name='',difftasks=['activity']):
    file_name = file_name.replace('.csv', '_pro.csv')
    my_df = pd.read_csv(file_name)
    hyper_paras_space = {'l2': hp.uniform('l2', 0, 0.01),
                         'dropout': hp.uniform('dropout', 0, 0.5),
                         'hidden_unit1': hp.choice('hidden_unit1', [64, 128, 256, 512]),
                         'hidden_unit2': hp.choice('hidden_unit2', [64, 128, 256, 512]),
                         'hidden_unit3': hp.choice('hidden_unit3', [64, 128, 256, 512])}

    reg = True if task_type == 'reg' else False
    epochs = 300  # training epoch

    batch_size = 128
    patience = 50
    opt_iters = 50

    #device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
    device = torch.device("cpu")
    # os.environ["CUDA_VISIBLE_DEVICES"] = gpu_id
    # if device == 'cuda':
    #    torch.cuda.set_device(eval(gpu_id))  # gpu device id
    args = {'device': device, 'metric': 'rmse' if reg else 'roc_auc', 'epochs': epochs,
            'patience': patience,  'reg': reg}
    random_state = 20



    while True:
        tr_indx, val_indx, te_indx = get_split_index(X, Y, split_type=split_type, random=random_state)
        data_tr_x, data_tr_y = my_df.loc[tr_indx,'Smiles'],my_df.loc[tr_indx,difftasks]
        data_va_x, data_va_y = my_df.loc[tr_indx,'Smiles'],my_df.loc[val_indx,difftasks]
        data_te_x, data_te_y = my_df.loc[tr_indx,'Smiles'],my_df.loc[te_indx,difftasks]


        data_tr_x, data_tr_y = create_des(data_tr_x, data_tr_y, FP_type=FP_type, model_dir=model_dir)
        data_va_x, data_va_y = create_des(data_va_x, data_va_y, FP_type=FP_type, model_dir=model_dir)
        data_te_x, data_te_y = create_des(data_te_x, data_te_y, FP_type=FP_type, model_dir=model_dir)

        if np.any(data_tr_y[difftasks].apply(all_one_zeros)) or \
                        np.any(data_va_y[difftasks].apply(all_one_zeros)) or \
                        np.any(data_te_y[difftasks].apply(all_one_zeros)):
            # print(
            #     '\ninvalid random seed {} due to one class presented in the {} splitted sets...'.format('None',
            #                                                                                             split_type))

            random_state += np.random.randint(50, 999999)

            # print('Changing to another random seed {}\n'.format(random_state))
        else:
            # print('random seed used in repetition {} is {}'.format(split_type, random_state))

            break
    data_tr_y = data_tr_y.values.reshape(-1,len(difftasks))
    data_va_y = data_va_y.values.reshape(-1,len(difftasks))
    data_te_y = data_te_y.values.reshape(-1,len(difftasks))

    # dataloader

    train_dataset = MyDataset(data_tr_x, data_tr_y)
    validation_dataset = MyDataset(data_va_x, data_va_y)
    test_dataset = MyDataset(data_te_x, data_te_y)
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    validation_loader = DataLoader(validation_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    inputs = data_tr_x.shape[1]
    if not reg:
        pos_weights = get_pos_weight(my_df[difftasks].values)


    def hyper_opt(hyper_paras):
        hidden_units = [hyper_paras['hidden_unit1'], hyper_paras['hidden_unit2'], hyper_paras['hidden_unit3']]
        my_model = MyDNN(inputs=inputs, hideen_units=hidden_units, dp_ratio=hyper_paras['dropout'],
                         outputs=len(difftasks), reg=reg)
        optimizer = torch.optim.Adadelta(my_model.parameters(), weight_decay=hyper_paras['l2'])
        if model_dir:
            file_nam = '%s/%s_%s_%s_%s_%.4f_%d_%d_%d_%.4f_early_stop.pth' % (model_dir,task_type,FP_type,split_type,'dataset', hyper_paras['dropout'],
                                                                  hyper_paras['hidden_unit1'],
                                                                  hyper_paras['hidden_unit2'],
                                                                  hyper_paras['hidden_unit3'],
                                                                  hyper_paras['l2'])
        if reg:
            loss_func = MSELoss(reduction='none')
            stopper = EarlyStopping(mode='lower', patience=patience, filename=file_nam)
        else:
            loss_func = BCEWithLogitsLoss(reduction='none', pos_weight=pos_weights.to(args['device']))
            stopper = EarlyStopping(mode='higher', patience=patience, filename=file_nam)
        my_model.to(device)
        for i in range(epochs):
            # training
            run_a_train_epoch(my_model, train_loader, loss_func, optimizer, args)

            # early stopping
            val_scores = run_an_eval_epoch(my_model, validation_loader, args)

            early_stop = stopper.step(val_scores[args['metric']], my_model)

            if early_stop:
                break
        stopper.load_checkpoint(my_model)
        val_scores = run_an_eval_epoch(my_model, validation_loader, args)
        feedback = val_scores[args['metric']] if reg else (1 - val_scores[args['metric']])

        my_model.cpu()
        torch.cuda.empty_cache()
        gc.collect()
        return feedback



    trials = Trials()  # 通过Trials捕获信息

    opt_res = fmin(hyper_opt, hyper_paras_space, algo=tpe.suggest, max_evals=opt_iters, trials=trials)



    # construct the model based on the optimal hyper-parameters
    hidden_unit1_ls = [64, 128, 256, 512]
    hidden_unit2_ls = [64, 128, 256, 512]
    hidden_unit3_ls = [64, 128, 256, 512]
    opt_hidden_units = [hidden_unit1_ls[opt_res['hidden_unit1']], hidden_unit2_ls[opt_res['hidden_unit2']],
                        hidden_unit3_ls[opt_res['hidden_unit3']]]
    best_model = MyDNN(inputs=inputs, hideen_units=opt_hidden_units, outputs=len(difftasks),
                       dp_ratio=opt_res['dropout'], reg=reg)

    best_file_name = '%s/%s_%s_%s_%s_%.4f_%d_%d_%d_%.4f_early_stop.pth' % (model_dir,task_type,FP_type,split_type,'dataset', opt_res['dropout'],
                                                                   hidden_unit1_ls[opt_res['hidden_unit1']],
                                                                   hidden_unit1_ls[opt_res['hidden_unit2']],
                                                                   hidden_unit1_ls[opt_res['hidden_unit3']],
                                                                   opt_res['l2'])

    best_model.load_state_dict(torch.load(best_file_name, map_location=device)['model_state_dict'])
    best_model.to(device)
    tr_scores = run_an_eval_epoch(best_model, train_loader, args)
    val_scores = run_an_eval_epoch(best_model, validation_loader, args)
    te_scores = run_an_eval_epoch(best_model, test_loader, args)
    pd_res = []
    if not reg :
        tr_results = [FP_type, split_type, 'tr', opt_hidden_units,opt_res['l2'],opt_res['dropout'], tr_scores['se'], tr_scores['sp'],
                      tr_scores['acc'], tr_scores['mcc'], tr_scores['pre'],
                      tr_scores['prc_auc'], tr_scores['roc_auc']]
        va_results = [FP_type, split_type, 'va', opt_hidden_units,opt_res['l2'],opt_res['dropout'], val_scores['se'], val_scores['sp'],
                      val_scores['acc'], val_scores['mcc'], val_scores['pre'],
                      val_scores['prc_auc'], val_scores['roc_auc']]
        te_results = [FP_type, split_type, 'te', opt_hidden_units,opt_res['l2'],opt_res['dropout'], te_scores['se'], te_scores['sp'],
                      te_scores['acc'], te_scores['mcc'], te_scores['pre'],
                      te_scores['prc_auc'], te_scores['roc_auc']]
        pd_res.append(tr_results)
        pd_res.append(va_results)
        pd_res.append(te_results)
        para_res = pd.DataFrame(pd_res, columns=['FP_type', 'split_type', 'type',
                    'opt_hidden_units', 'l2','dropout','se', 'sp','acc', 'mcc', 'precision','auc_prc', 'auc_roc'])
    else :
        tr_results = [FP_type, split_type, 'tr',
                      opt_hidden_units, opt_res['l2'], opt_res['dropout'],  tr_scores['rmse'],tr_scores['mae'], tr_scores['r2']]
        va_results = [FP_type, split_type, 'va',
                      opt_hidden_units, opt_res['l2'], opt_res['dropout'],  val_scores['rmse'],val_scores['mae'], val_scores['r2']]
        te_results = [FP_type, split_type, 'te',
                      opt_hidden_units, opt_res['l2'], opt_res['dropout'],  te_scores['rmse'],te_scores['mae'], te_scores['r2']]
        pd_res.append(tr_results)
        pd_res.append(va_results)
        pd_res.append(te_results)
        para_res = pd.DataFrame(pd_res, columns=['FP_type', 'split_type', 'type',
                                                 'opt_hidden_units','l2','dropout',
                                                 'rmse','mae','r2'])


    pd_res = []
    for i in range(10):
        item = best_model_runing((i + 1) * 500,opt_res, X, Y, split_type=split_type, FP_type=FP_type,task_type=task_type,model_dir=model_dir,difftasks=difftasks,my_df=my_df)
        pd_res.extend(item)
    if  not reg :

        best_res = pd.DataFrame(pd_res, columns=['seed', 'FP_type', 'split_type', 'type',
                                                  'se', 'sp',
                                                 'acc', 'mcc','precision', 'auc_prc', 'auc_roc'])
    else:
        best_res = pd.DataFrame(pd_res, columns=['seed', 'FP_type', 'split_type', 'type',
                                                 'rmse','mae','r2'])

    return para_res,best_res





import pandas as pd
import sklearn
from sklearn.model_selection import train_test_split
import deepchem as dc
from rdkit import Chem
from rdkit import DataStructs
from rdkit.ML.Cluster import Butina
from rdkit.Chem import rdFingerprintGenerator
def tanimoto_distance_matrix(fp_list):
    """Calculate distance matrix for fingerprint list"""
    dissimilarity_matrix = []

    for i in range(1, len(fp_list)):
        # Compare the current fingerprint against all the previous ones in the list
        similarities = DataStructs.BulkTanimotoSimilarity(fp_list[i], fp_list[:i])
        # Since we need a distance matrix, calculate 1-x for every element in similarity matrix
        dissimilarity_matrix.extend([1 - x for x in similarities])
    return dissimilarity_matrix
def cluster_fingerprints(fingerprints, cutoff=0.2):
    """Cluster fingerprints
    Parameters:
        fingerprints
        cutoff: threshold for the clustering
    """
    # Calculate Tanimoto distance matrix
    distance_matrix = tanimoto_distance_matrix(fingerprints)
    # Now cluster the data with the implemented Butina algorithm:
    clusters = Butina.ClusterData(distance_matrix, len(fingerprints), cutoff, isDistData=True)
    clusters = sorted(clusters, key=len, reverse=True)
    return clusters
def split_dataset(X,Y,split_type='random',valid_need = False,train_size=0.80,random_state=20):
    if valid_need is True:
        if split_type == 'random':
            X_train, X_rest, Y_train, Y_rest = train_test_split(X, Y, train_size=train_size, random_state=random_state)
            X_valid, X_test, Y_valid, Y_test = train_test_split(X_rest, Y_rest, train_size=0.5, random_state=random_state)
        elif split_type == 'scaffold':
            chembl_data = pd.concat([X, Y], axis=1)
            chembl_data.columns = ['Smiles', 'activity']
            with dc.utils.UniversalNamedTemporaryFile(mode='w') as tmpfile:
                chembl_data.to_csv(tmpfile.name)
                featurizer = dc.feat.CircularFingerprint(size=1024)
                loader = dc.data.CSVLoader(["activity"], feature_field="Smiles", featurizer=featurizer)
                active = loader.create_dataset(tmpfile.name)
                scaffoldsplitter = dc.splits.ScaffoldSplitter()
                train, rest = scaffoldsplitter.train_test_split(active,seed=random_state,frac_train=train_size)
                train_df = pd.DataFrame(train.ids, columns=['Smiles'])
                rest_df = pd.DataFrame(rest.ids, columns=['Smiles'])
                train = chembl_data[chembl_data['Smiles'].isin(train_df['Smiles'].tolist())]
                rest = chembl_data[chembl_data['Smiles'].isin(rest_df['Smiles'].tolist())]
                X_train = train.Smiles
                X_rest = rest.Smiles
                Y_train = train['activity']
                Y_rest = rest['activity']
                X_valid, X_test, Y_valid, Y_test = train_test_split(X_rest, Y_rest, train_size=0.5,
                                                                    random_state=random_state)

        elif split_type == 'cluster':
            compound_df = pd.concat([X, Y], axis=1)
            compound_df.columns = ['Smiles', 'activity']
            compounds = []
            for _, chembl_id, smiles in compound_df[["activity", "Smiles"]].itertuples():
                compounds.append((Chem.MolFromSmiles(smiles), chembl_id))
            rdkit_gen = rdFingerprintGenerator.GetRDKitFPGenerator(maxPath=5)
            fingerprints = [rdkit_gen.GetFingerprint(mol) for mol, idx in compounds]
            clusters = cluster_fingerprints(fingerprints, cutoff=0.2)
            cluster_centers = [compounds[c[0]] for c in clusters]
            sorted_clusters = []
            Singletons = []
            for cluster in clusters:
                if len(cluster) <= 1:
                    Singletons.append(cluster)
                    continue  # Singletons
                sorted_fingerprints = [rdkit_gen.GetFingerprint(compounds[i][0]) for i in cluster]
                similarities = DataStructs.BulkTanimotoSimilarity(
                    sorted_fingerprints[0], sorted_fingerprints[1:]
                )
                similarities = list(zip(similarities, cluster[1:]))
                similarities.sort(reverse=True)
                sorted_clusters.append((len(similarities), [i for _, i in similarities]))
                sorted_clusters.sort(reverse=True)
            selected_molecules = cluster_centers.copy()
            index = 0
            pending = int(len(compounds) * train_size) - len(selected_molecules)
            while pending > 0 and index < len(sorted_clusters):
                tmp_cluster = sorted_clusters[index][1]
                if sorted_clusters[index][0] > 10:
                    num_compounds = int(sorted_clusters[index][0] * train_size)
                else:
                    num_compounds = len(tmp_cluster)
                if num_compounds > pending:
                    num_compounds = pending
                selected_molecules += [compounds[i] for i in tmp_cluster[:num_compounds]]
                index += 1
                pending = int(len(compounds) * train_size) - len(selected_molecules)
            test = [i for i in compounds if i not in selected_molecules]

            train = selected_molecules
            X_train = pd.DataFrame(train, columns=['Smiles', 'activity'])['Smiles']
            X_train = X_train.apply(lambda x: Chem.MolToSmiles(x))
            Y_train = pd.DataFrame(train, columns=['Smiles', 'activity'])['activity']
            X_rest = pd.DataFrame(test, columns=['Smiles', 'activity'])
            X_rest = sklearn.utils.shuffle(X_rest, random_state=random_state)
            X_valid = X_rest.iloc[:int(len(X_rest) / 2), :]['Smiles']
            X_valid = X_valid.apply(lambda x: Chem.MolToSmiles(x))
            X_test = X_rest.iloc[int(len(X_rest) / 2):, :]['Smiles']
            X_test = X_test.apply(lambda x: Chem.MolToSmiles(x))
            Y_valid = X_rest.iloc[:int(len(X_rest) / 2), :]['activity']
            Y_test = X_rest.iloc[int(len(X_rest) / 2):, :]['activity']
        return X_train,X_valid,X_test,Y_train,Y_valid,Y_test
    else:
        if split_type == 'random':
            X_train, X_test, Y_train, Y_test = train_test_split(X, Y, train_size = train_size, random_state=random_state)
        elif split_type == 'scaffold':
            chembl_data = pd.concat([X, Y], axis=1)
            chembl_data.columns = ['Smiles', 'activity']
            with dc.utils.UniversalNamedTemporaryFile(mode='w') as tmpfile:
                chembl_data.to_csv(tmpfile.name)
                featurizer = dc.feat.CircularFingerprint(size=1024)
                loader = dc.data.CSVLoader(["activity"], feature_field="Smiles", featurizer=featurizer)
                active = loader.create_dataset(tmpfile.name)
                scaffoldsplitter = dc.splits.ScaffoldSplitter()
                train, test = scaffoldsplitter.train_test_split(active,frac_train = train_size,seed=random_state)
                train_df = pd.DataFrame(train.ids, columns=['Smiles'])
                test_df = pd.DataFrame(test.ids, columns=['Smiles'])
                train = chembl_data[chembl_data['Smiles'].isin(train_df['Smiles'].tolist())]
                test = chembl_data[chembl_data['Smiles'].isin(test_df['Smiles'].tolist())]
                X_train = train.Smiles
                X_test = test.Smiles
                Y_train = train.activity
                Y_test = test.activity
        elif split_type == 'cluster':
            compound_df = pd.concat([X, Y], axis=1)
            compound_df.columns = ['Smiles', 'activity']
            compound_df['inter_smiles'] = compound_df['Smiles'].apply(lambda x: Chem.MolToSmiles(Chem.MolFromSmiles(x)))
            compounds = []
            for _, chembl_id, smiles in compound_df[["activity", "inter_smiles"]].itertuples():
                compounds.append((Chem.MolFromSmiles(smiles), chembl_id))
            rdkit_gen = rdFingerprintGenerator.GetRDKitFPGenerator(maxPath=5)
            fingerprints = [rdkit_gen.GetFingerprint(mol) for mol, idx in compounds]
            clusters = cluster_fingerprints(fingerprints, cutoff=0.2)
            cluster_centers = [compounds[c[0]] for c in clusters]
            sorted_clusters = []
            Singletons = []
            for cluster in clusters:
                if len(cluster) <= 1:
                    Singletons.append(cluster)
                    continue  # Singletons
                sorted_fingerprints = [rdkit_gen.GetFingerprint(compounds[i][0]) for i in cluster]
                similarities = DataStructs.BulkTanimotoSimilarity(
                    sorted_fingerprints[0], sorted_fingerprints[1:]
                )
                similarities = list(zip(similarities, cluster[1:]))
                similarities.sort(reverse=True)
                sorted_clusters.append((len(similarities), [i for _, i in similarities]))
                sorted_clusters.sort(reverse=True)
            selected_molecules = cluster_centers.copy()
            index = 0
            pending = int(len(compounds) * train_size) - len(selected_molecules)
            while pending > 0 and index < len(sorted_clusters):
                tmp_cluster = sorted_clusters[index][1]
                if sorted_clusters[index][0] > 10:
                    num_compounds = int(sorted_clusters[index][0] * train_size)
                else:
                    num_compounds = len(tmp_cluster)
                if num_compounds > pending:
                    num_compounds = pending
                selected_molecules += [compounds[i] for i in tmp_cluster[:num_compounds]]
                index += 1
                pending = int(len(compounds) * train_size) - len(selected_molecules)
            test = [i for i in compounds if i not in selected_molecules]
            train = selected_molecules
            X_train = pd.DataFrame(train, columns=['Smiles', 'activity'])['Smiles']
            X_train = X_train.apply(lambda x: Chem.MolToSmiles(x))
            Y_train = pd.DataFrame(train, columns=['Smiles', 'activity'])['activity']
            X_test = pd.DataFrame(test, columns=['Smiles', 'activity'])['Smiles']
            X_test = X_test.apply(lambda x: Chem.MolToSmiles(x))
            Y_test = pd.DataFrame(test, columns=['Smiles', 'activity'])['activity']
        return  X_train, X_test, Y_train, Y_test




